from enum import Enum
from typing import Optional, List, Dict, Type, Literal
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
from rich.console import Console
import time
import json
import os
from pydantic import BaseModel, ValidationError
import asyncio
import portalocker
import threading

from deepeval.prompt.api import (
    PromptHttpResponse,
    PromptMessage,
    PromptType,
    PromptInterpolationType,
    PromptPushRequest,
    PromptVersionsHttpResponse,
    PromptMessageList,
    PromptUpdateRequest,
    ModelSettings,
    OutputSchema,
    OutputType,
    ReasoningEffort,
    Verbosity,
    ModelProvider,
)
from deepeval.prompt.utils import (
    interpolate_text,
    construct_base_model,
    construct_output_schema,
)
from deepeval.confident.api import Api, Endpoints, HttpMethods
from deepeval.constants import HIDDEN_DIR

CACHE_FILE_NAME = f"{HIDDEN_DIR}/.deepeval-prompt-cache.json"
VERSION_CACHE_KEY = "version"
LABEL_CACHE_KEY = "label"

# Global background event loop for polling
_polling_loop: Optional[asyncio.AbstractEventLoop] = None
_polling_thread: Optional[threading.Thread] = None
_polling_loop_lock = threading.Lock()


def _get_or_create_polling_loop() -> asyncio.AbstractEventLoop:
    """Get or create a background event loop for polling that runs in a daemon thread."""
    global _polling_loop, _polling_thread

    with _polling_loop_lock:
        if _polling_loop is None or not _polling_loop.is_running():

            def run_loop():
                global _polling_loop
                _polling_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(_polling_loop)
                _polling_loop.run_forever()

            _polling_thread = threading.Thread(target=run_loop, daemon=True)
            _polling_thread.start()

            # Wait for loop to be ready
            while _polling_loop is None:
                time.sleep(0.01)

        return _polling_loop


class CustomEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Enum):
            return obj.value
        elif isinstance(obj, BaseModel):
            return obj.model_dump(by_alias=True, exclude_none=True)
        return json.JSONEncoder.default(self, obj)


class CachedPrompt(BaseModel):
    alias: str
    version: str
    label: Optional[str] = None
    template: Optional[str]
    messages_template: Optional[List[PromptMessage]]
    prompt_version_id: str
    type: PromptType
    interpolation_type: PromptInterpolationType
    model_settings: Optional[ModelSettings]
    output_type: Optional[OutputType]
    output_schema: Optional[OutputSchema]

    class Config:
        use_enum_values = True


class Prompt:

    def __init__(
        self,
        alias: Optional[str] = None,
        text_template: Optional[str] = None,
        messages_template: Optional[List[PromptMessage]] = None,
        model_settings: Optional[ModelSettings] = None,
        output_type: Optional[OutputType] = None,
        output_schema: Optional[Type[BaseModel]] = None,
    ):
        if text_template and messages_template:
            raise TypeError(
                "Unable to create Prompt where 'text_template' and 'messages_template' are both provided. Please provide only one to continue."
            )
        self.alias = alias
        self.text_template = text_template
        self.messages_template = messages_template
        self.model_settings: Optional[ModelSettings] = model_settings
        self.output_type: Optional[OutputType] = output_type
        self.output_schema: Optional[Type[BaseModel]] = output_schema
        self.label: Optional[str] = None
        self.interpolation_type: Optional[PromptInterpolationType] = None

        self._version = None
        self._prompt_version_id: Optional[str] = None
        self._polling_tasks: Dict[str, Dict[str, asyncio.Task]] = {}
        self._refresh_map: Dict[str, Dict[str, int]] = {}
        self._lock = (
            threading.Lock()
        )  # Protect instance attributes from race conditions

        self.type: Optional[PromptType] = None
        if text_template:
            self.type = PromptType.TEXT
        elif messages_template:
            self.type = PromptType.LIST

    def __del__(self):
        """Cleanup polling tasks when instance is destroyed"""
        try:
            self._stop_polling()
        except Exception:
            # Suppress exceptions during cleanup to avoid issues in interpreter shutdown
            pass

    @property
    def version(self):
        if self._version is not None and self._version != "latest":
            return self._version
        versions = self._get_versions()
        if len(versions) == 0:
            return "latest"
        else:
            return versions[-1].version

    @version.setter
    def version(self, value):
        self._version = value

    def load(self, file_path: str, messages_key: Optional[str] = None):
        _, ext = os.path.splitext(file_path)
        if ext != ".json" and ext != ".txt":
            raise ValueError("Only .json and .txt files are supported")

        file_name = os.path.basename(file_path).split(".")[0]
        self.alias = file_name
        with open(file_path, "r") as f:
            content = f.read()
        try:
            data = json.loads(content)
        except:
            self.text_template = content
            return content

        text_template = None
        messages_template = None
        try:
            if isinstance(data, list):
                messages_template = PromptMessageList.validate_python(data)
            elif isinstance(data, dict):
                if messages_key is None:
                    raise ValueError(
                        "messages `key` must be provided if file is a dictionary"
                    )
                messages = data[messages_key]
                messages_template = PromptMessageList.validate_python(messages)
            else:
                text_template = content
        except ValidationError:
            text_template = content

        self.text_template = text_template
        self.messages_template = messages_template
        return text_template or messages_template

    def interpolate(self, **kwargs):
        with self._lock:
            prompt_type = self.type
            text_template = self.text_template
            messages_template = self.messages_template
            interpolation_type = self.interpolation_type

        if prompt_type == PromptType.TEXT:
            if text_template is None:
                raise TypeError(
                    "Unable to interpolate empty prompt template. Please pull a prompt from Confident AI or set template manually to continue."
                )

            return interpolate_text(interpolation_type, text_template, **kwargs)

        elif prompt_type == PromptType.LIST:
            if messages_template is None:
                raise TypeError(
                    "Unable to interpolate empty prompt template messages. Please pull a prompt from Confident AI or set template manually to continue."
                )

            interpolated_messages = []
            for message in messages_template:
                interpolated_content = interpolate_text(
                    interpolation_type, message.content, **kwargs
                )
                interpolated_messages.append(
                    {"role": message.role, "content": interpolated_content}
                )
            return interpolated_messages
        else:
            raise ValueError(f"Unsupported prompt type: {self.type}")

    ############################################
    ### Utils
    ############################################

    def _get_versions(self) -> List:
        if self.alias is None:
            raise ValueError(
                "Prompt alias is not set. Please set an alias to continue."
            )
        api = Api()
        data, _ = api.send_request(
            method=HttpMethods.GET,
            endpoint=Endpoints.PROMPTS_VERSIONS_ENDPOINT,
            url_params={"alias": self.alias},
        )
        versions = PromptVersionsHttpResponse(**data)
        return versions.text_versions or versions.messages_versions or []

    def _read_from_cache(
        self,
        alias: str,
        version: Optional[str] = None,
        label: Optional[str] = None,
    ) -> Optional[CachedPrompt]:
        if not os.path.exists(CACHE_FILE_NAME):
            return None

        try:
            # Use shared lock for reading to allow concurrent reads
            with portalocker.Lock(
                CACHE_FILE_NAME,
                mode="r",
                flags=portalocker.LOCK_SH | portalocker.LOCK_NB,
            ) as f:
                cache_data = json.load(f)

            if alias in cache_data:
                if version:
                    if (
                        VERSION_CACHE_KEY in cache_data[alias]
                        and version in cache_data[alias][VERSION_CACHE_KEY]
                    ):
                        return CachedPrompt(
                            **cache_data[alias][VERSION_CACHE_KEY][version]
                        )
                elif label:
                    if (
                        LABEL_CACHE_KEY in cache_data[alias]
                        and label in cache_data[alias][LABEL_CACHE_KEY]
                    ):
                        return CachedPrompt(
                            **cache_data[alias][LABEL_CACHE_KEY][label]
                        )
            return None
        except (portalocker.exceptions.LockException, Exception):
            # If cache is locked, corrupted or unreadable, return None and let it fetch from API
            return None

    def _write_to_cache(
        self,
        cache_key: Literal[VERSION_CACHE_KEY, LABEL_CACHE_KEY],
        version: str,
        label: Optional[str] = None,
        text_template: Optional[str] = None,
        messages_template: Optional[List[PromptMessage]] = None,
        prompt_version_id: Optional[str] = None,
        type: Optional[PromptType] = None,
        interpolation_type: Optional[PromptInterpolationType] = None,
        model_settings: Optional[ModelSettings] = None,
        output_type: Optional[OutputType] = None,
        output_schema: Optional[OutputSchema] = None,
    ):
        if not self.alias:
            return

        # Ensure directory exists
        os.makedirs(HIDDEN_DIR, exist_ok=True)

        try:
            # Use r+ mode if file exists, w mode if it doesn't
            mode = "r+" if os.path.exists(CACHE_FILE_NAME) else "w"

            with portalocker.Lock(
                CACHE_FILE_NAME,
                mode=mode,
                flags=portalocker.LOCK_EX,
            ) as f:
                # Read existing cache data if file exists and has content
                cache_data = {}
                if mode == "r+":
                    try:
                        f.seek(0)
                        content = f.read()
                        if content:
                            cache_data = json.loads(content)
                    except (json.JSONDecodeError, Exception):
                        cache_data = {}

                # Ensure the cache structure is initialized properly
                if self.alias not in cache_data:
                    cache_data[self.alias] = {}

                if cache_key not in cache_data[self.alias]:
                    cache_data[self.alias][cache_key] = {}

                # Cache the prompt
                cached_entry = {
                    "alias": self.alias,
                    "version": version,
                    "label": label,
                    "template": text_template,
                    "messages_template": messages_template,
                    "prompt_version_id": prompt_version_id,
                    "type": type,
                    "interpolation_type": interpolation_type,
                    "model_settings": model_settings,
                    "output_type": output_type,
                    "output_schema": output_schema,
                }

                if cache_key == VERSION_CACHE_KEY:
                    cache_data[self.alias][cache_key][version] = cached_entry
                else:
                    cache_data[self.alias][cache_key][label] = cached_entry

                # Write back to cache file
                f.seek(0)
                f.truncate()
                json.dump(cache_data, f, cls=CustomEncoder)
        except portalocker.exceptions.LockException:
            # If we can't acquire the lock, silently skip caching
            pass
        except Exception:
            # If any other error occurs during caching, silently skip
            pass

    def _load_from_cache_with_progress(
        self,
        progress: Progress,
        task_id: int,
        start_time: float,
        version: Optional[str] = None,
        label: Optional[str] = None,
    ):
        """
        Load prompt from cache and update progress bar.
        Raises if unable to load from cache.
        """
        cached_prompt = self._read_from_cache(
            self.alias, version=version, label=label
        )
        if not cached_prompt:
            raise ValueError("Unable to fetch prompt and load from cache")

        with self._lock:
            self._version = cached_prompt.version
            self.label = cached_prompt.label
            self.text_template = cached_prompt.template
            self.messages_template = cached_prompt.messages_template
            self._prompt_version_id = cached_prompt.prompt_version_id
            self.type = (
                PromptType(cached_prompt.type) if cached_prompt.type else None
            )
            self.interpolation_type = (
                PromptInterpolationType(cached_prompt.interpolation_type)
                if cached_prompt.interpolation_type
                else None
            )
            self.model_settings = cached_prompt.model_settings
            self.output_type = (
                OutputType(cached_prompt.output_type)
                if cached_prompt.output_type
                else None
            )
            self.output_schema = construct_base_model(
                cached_prompt.output_schema
            )

        end_time = time.perf_counter()
        time_taken = format(end_time - start_time, ".2f")
        progress.update(
            task_id,
            description=f"{progress.tasks[task_id].description}[rgb(25,227,160)]Loaded from cache! ({time_taken}s)",
        )

    ############################################
    ### Pull, Push, Update
    ############################################

    def pull(
        self,
        version: Optional[str] = None,
        label: Optional[str] = None,
        fallback_to_cache: bool = True,
        write_to_cache: bool = True,
        default_to_cache: bool = True,
        refresh: Optional[int] = 60,
    ):
        should_write_on_first_fetch = False
        if refresh:
            # Check if we need to bootstrap the cache
            cached_prompt = self._read_from_cache(
                self.alias, version=version, label=label
            )
            if cached_prompt is None:
                # No cache exists, so we should write after fetching to bootstrap
                should_write_on_first_fetch = True
            write_to_cache = False  # Polling will handle subsequent writes

        if self.alias is None:
            raise TypeError(
                "Unable to pull prompt from Confident AI when no alias is provided."
            )

        # Manage background prompt polling
        if refresh:
            loop = _get_or_create_polling_loop()
            asyncio.run_coroutine_threadsafe(
                self.create_polling_task(version, label, refresh), loop
            )

        if default_to_cache:
            try:
                cached_prompt = self._read_from_cache(
                    self.alias, version=version, label=label
                )
                if cached_prompt:
                    with self._lock:
                        self._version = cached_prompt.version
                        self.label = cached_prompt.label
                        self.text_template = cached_prompt.template
                        self.messages_template = cached_prompt.messages_template
                        self._prompt_version_id = (
                            cached_prompt.prompt_version_id
                        )
                        self.type = (
                            PromptType(cached_prompt.type)
                            if cached_prompt.type
                            else None
                        )
                        self.interpolation_type = (
                            PromptInterpolationType(
                                cached_prompt.interpolation_type
                            )
                            if cached_prompt.interpolation_type
                            else None
                        )
                        self.model_settings = cached_prompt.model_settings
                        self.output_type = (
                            OutputType(cached_prompt.output_type)
                            if cached_prompt.output_type
                            else None
                        )
                        self.output_schema = construct_base_model(
                            cached_prompt.output_schema
                        )
                    return
            except:
                pass

        api = Api()
        with Progress(
            SpinnerColumn(style="rgb(106,0,255)"),
            BarColumn(bar_width=60),
            TextColumn("[progress.description]{task.description}"),
            transient=False,
        ) as progress:
            HINT_TEXT = (
                f"version='{version or 'latest'}'"
                if not label
                else f"label='{label}'"
            )
            task_id = progress.add_task(
                f"Pulling [rgb(106,0,255)]'{self.alias}' ({HINT_TEXT})[/rgb(106,0,255)] from Confident AI...",
                total=100,
            )

            start_time = time.perf_counter()
            try:
                if label:
                    data, _ = api.send_request(
                        method=HttpMethods.GET,
                        endpoint=Endpoints.PROMPTS_LABEL_ENDPOINT,
                        url_params={
                            "alias": self.alias,
                            "label": label,
                        },
                    )
                else:
                    data, _ = api.send_request(
                        method=HttpMethods.GET,
                        endpoint=Endpoints.PROMPTS_VERSION_ID_ENDPOINT,
                        url_params={
                            "alias": self.alias,
                            "versionId": version or "latest",
                        },
                    )

                response = PromptHttpResponse(
                    id=data["id"],
                    version=data.get("version", None),
                    label=data.get("label", None),
                    text=data.get("text", None),
                    messages=data.get("messages", None),
                    type=data["type"],
                    interpolation_type=data["interpolationType"],
                    model_settings=data.get("modelSettings", None),
                    output_type=data.get("outputType", None),
                    output_schema=data.get("outputSchema", None),
                )
            except Exception:
                if fallback_to_cache:
                    self._load_from_cache_with_progress(
                        progress,
                        task_id,
                        start_time,
                        version=version,
                        label=label,
                    )
                    return
                raise

            with self._lock:
                self._version = response.version
                self.label = response.label
                self.text_template = response.text
                self.messages_template = response.messages
                self._prompt_version_id = response.id
                self.type = response.type
                self.interpolation_type = response.interpolation_type
                self.model_settings = response.model_settings
                self.output_type = response.output_type
                self.output_schema = construct_base_model(
                    response.output_schema
                )

            end_time = time.perf_counter()
            time_taken = format(end_time - start_time, ".2f")
            progress.update(
                task_id,
                description=f"{progress.tasks[task_id].description}[rgb(25,227,160)]Done! ({time_taken}s)",
            )
            # Write to cache if explicitly requested OR if we need to bootstrap cache for refresh mode
            if write_to_cache or should_write_on_first_fetch:
                self._write_to_cache(
                    cache_key=LABEL_CACHE_KEY if label else VERSION_CACHE_KEY,
                    version=response.version,
                    label=response.label,
                    text_template=response.text,
                    messages_template=response.messages,
                    prompt_version_id=response.id,
                    type=response.type,
                    interpolation_type=response.interpolation_type,
                    model_settings=response.model_settings,
                    output_type=response.output_type,
                    output_schema=response.output_schema,
                )

    def push(
        self,
        text: Optional[str] = None,
        messages: Optional[List[PromptMessage]] = None,
        interpolation_type: Optional[
            PromptInterpolationType
        ] = PromptInterpolationType.FSTRING,
        model_settings: Optional[ModelSettings] = None,
        output_type: Optional[OutputType] = None,
        output_schema: Optional[Type[BaseModel]] = None,
        _verbose: Optional[bool] = True,
    ):
        if self.alias is None:
            raise ValueError(
                "Prompt alias is not set. Please set an alias to continue."
            )
        text_template = text or self.text_template
        messages_template = messages or self.messages_template
        if text_template is None and messages_template is None:
            raise ValueError("Either text or messages must be provided")
        if text_template is not None and messages_template is not None:
            raise ValueError("Only one of text or messages can be provided")

        body = PromptPushRequest(
            alias=self.alias,
            text=text_template,
            messages=messages_template,
            interpolation_type=interpolation_type or self.interpolation_type,
            model_settings=model_settings or self.model_settings,
            output_type=output_type or self.output_type,
            output_schema=construct_output_schema(output_schema)
            or construct_output_schema(self.output_schema),
        )
        try:
            body = body.model_dump(
                by_alias=True, exclude_none=True, mode="json"
            )
        except AttributeError:
            # Pydantic version below 2.0
            body = body.dict(by_alias=True, exclude_none=True)

        api = Api()
        _, link = api.send_request(
            method=HttpMethods.POST,
            endpoint=Endpoints.PROMPTS_ENDPOINT,
            body=body,
        )
        versions = self._get_versions()

        if link and versions:
            self._prompt_version_id = versions[-1].id
            self.text_template = text_template
            self.messages_template = messages_template
            self.interpolation_type = (
                interpolation_type or self.interpolation_type
            )
            self.model_settings = model_settings or self.model_settings
            self.output_type = output_type or self.output_type
            self.output_schema = output_schema or self.output_schema
            self.type = PromptType.TEXT if text_template else PromptType.LIST
            if _verbose:
                console = Console()
                console.print(
                    "✅ Prompt successfully pushed to Confident AI! View at "
                    f"[link={link}]{link}[/link]"
                )

    def update(
        self,
        version: str,
        text: Optional[str] = None,
        messages: Optional[List[PromptMessage]] = None,
        interpolation_type: Optional[
            PromptInterpolationType
        ] = PromptInterpolationType.FSTRING,
        model_settings: Optional[ModelSettings] = None,
        output_type: Optional[OutputType] = None,
        output_schema: Optional[Type[BaseModel]] = None,
    ):
        if self.alias is None:
            raise ValueError(
                "Prompt alias is not set. Please set an alias to continue."
            )

        body = PromptUpdateRequest(
            text=text,
            messages=messages,
            interpolation_type=interpolation_type,
            model_settings=model_settings,
            output_type=output_type,
            output_schema=construct_output_schema(output_schema),
        )
        try:
            body = body.model_dump(
                by_alias=True, exclude_none=True, mode="json"
            )
        except AttributeError:
            body = body.dict(by_alias=True, exclude_none=True)
        api = Api()
        data, _ = api.send_request(
            method=HttpMethods.PUT,
            endpoint=Endpoints.PROMPTS_VERSION_ID_ENDPOINT,
            url_params={
                "alias": self.alias,
                "versionId": version,
            },
            body=body,
        )
        if data:
            self._version = version
            self.text_template = text
            self.messages_template = messages
            self.interpolation_type = interpolation_type
            self.model_settings = model_settings
            self.output_type = output_type
            self.output_schema = output_schema
            self.type = PromptType.TEXT if text else PromptType.LIST
            console = Console()
            console.print("✅ Prompt successfully updated on Confident AI!")

    ############################################
    ### Polling
    ############################################

    async def create_polling_task(
        self,
        version: Optional[str],
        label: Optional[str],
        refresh: Optional[int] = 60,
    ):
        # If polling task doesn't exist, start it
        CACHE_KEY = LABEL_CACHE_KEY if label else VERSION_CACHE_KEY
        cache_value = label if label else version

        # Initialize nested dicts if they don't exist
        if CACHE_KEY not in self._polling_tasks:
            self._polling_tasks[CACHE_KEY] = {}
        if CACHE_KEY not in self._refresh_map:
            self._refresh_map[CACHE_KEY] = {}

        polling_task: Optional[asyncio.Task] = self._polling_tasks[
            CACHE_KEY
        ].get(cache_value)

        if refresh:
            self._refresh_map[CACHE_KEY][cache_value] = refresh
            if not polling_task:
                self._polling_tasks[CACHE_KEY][cache_value] = (
                    asyncio.create_task(self.poll(version, label))
                )

        # If invalid `refresh`, stop the task
        else:
            if polling_task:
                polling_task.cancel()
            if cache_value in self._polling_tasks[CACHE_KEY]:
                self._polling_tasks[CACHE_KEY].pop(cache_value)
            if cache_value in self._refresh_map[CACHE_KEY]:
                self._refresh_map[CACHE_KEY].pop(cache_value)

    async def poll(
        self,
        version: Optional[str] = None,
        label: Optional[str] = None,
    ):
        CACHE_KEY = LABEL_CACHE_KEY if label else VERSION_CACHE_KEY
        cache_value = label if label else version

        while True:
            await asyncio.sleep(self._refresh_map[CACHE_KEY][cache_value])

            api = Api()
            try:
                if label:
                    data, _ = api.send_request(
                        method=HttpMethods.GET,
                        endpoint=Endpoints.PROMPTS_LABEL_ENDPOINT,
                        url_params={
                            "alias": self.alias,
                            "label": label,
                        },
                    )
                else:
                    data, _ = api.send_request(
                        method=HttpMethods.GET,
                        endpoint=Endpoints.PROMPTS_VERSION_ID_ENDPOINT,
                        url_params={
                            "alias": self.alias,
                            "versionId": version or "latest",
                        },
                    )

                response = PromptHttpResponse(
                    id=data["id"],
                    version=data.get("version", None),
                    label=data.get("label", None),
                    text=data.get("text", None),
                    messages=data.get("messages", None),
                    type=data["type"],
                    interpolation_type=data["interpolationType"],
                )

                # Update the cache with fresh data from server
                self._write_to_cache(
                    cache_key=CACHE_KEY,
                    version=response.version,
                    label=response.label,
                    text_template=response.text,
                    messages_template=response.messages,
                    prompt_version_id=response.id,
                    type=response.type,
                    interpolation_type=response.interpolation_type,
                )

                # Update in-memory properties with fresh data (thread-safe)
                with self._lock:
                    self._version = response.version
                    self.label = response.label
                    self.text_template = response.text
                    self.messages_template = response.messages
                    self._prompt_version_id = response.id
                    self.type = response.type
                    self.interpolation_type = response.interpolation_type

            except Exception:
                pass

    def _stop_polling(self):
        loop = _polling_loop
        if not loop or not loop.is_running():
            return

        # Stop all polling tasks
        for ck in list(self._polling_tasks.keys()):
            for cv in list(self._polling_tasks[ck].keys()):
                task = self._polling_tasks[ck][cv]
                if task and not task.done():
                    loop.call_soon_threadsafe(task.cancel)
            self._polling_tasks[ck].clear()
            self._refresh_map[ck].clear()
        return



from pydantic import BaseModel, ConfigDict
from typing import Any, Optional


class ApiResponse(BaseModel):
    model_config = ConfigDict(extra="ignore")

    success: bool
    data: Optional[Any] = None
    error: Optional[str] = None
    deprecated: Optional[bool] = None
    link: Optional[str] = None


class ConfidentApiError(Exception):
    """Custom exception that preserves API response metadata"""

    def __init__(self, message: str, link: Optional[str] = None):
        super().__init__(message)
        self.link = link



from typing import List
import textwrap
import json

from deepeval.dataset import ConversationalGolden
from deepeval.test_case import Turn


class ConversationSimulatorTemplate:

    @staticmethod
    def simulate_first_user_turn(
        golden: ConversationalGolden, language: str
    ) -> str:
        prompt = textwrap.dedent(
            f"""Pretend you are a user of an LLM app. Your goal is to start a conversation in {language} based on a scenario 
            and user profile. The scenario defines your context and motivation for interacting with the LLM, 
            while the user profile provides additional personal details to make the conversation realistic and relevant.

            Guidelines:
            1. The opening message should clearly convey the user's intent or need within the scenario.
            2. Keep the tone warm, conversational, and natural, as if it’s from a real person seeking assistance.
            3. Avoid providing excessive details upfront; the goal is to initiate the conversation and build rapport, not to solve it in the first message.
            4. The message should be concise, ideally no more than 1-3 sentences.

            IMPORTANT: The output must be formatted as a JSON object with a single key `simulated_input`, where the value is the generated opening message in {language}.

            Example Language: english
            Example User Profile: "Jeff Seid, is available Monday and Thursday afternoons, and their phone number is 0010281839. He suffers from chronic migraines."
            Example Scenario: "A sick person trying to get a diagnosis for persistent headaches and fever."
            Example JSON Output:
            {{
                "simulated_input": "Hi, I haven’t been feeling well lately. I’ve had these headaches and a fever that just won’t go away. Could you help me figure out what’s going on?"
            }}

            Language: {language}
            User Profile: "{golden.user_description}"             
            Scenario: "{golden.scenario}"
            JSON Output:
        """
        )
        return prompt

    @staticmethod
    def simulate_user_turn(
        golden: ConversationalGolden,
        turns: List[Turn],
        language: str,
    ) -> str:
        previous_conversation = json.dumps(
            [t.model_dump() for t in turns], indent=4
        )
        prompt = textwrap.dedent(
            f"""
            Pretend you are a user of an LLM app. Your task is to generate the next user input in {language} 
            based on the provided scenario, user profile, and the previous conversation.

            Guidelines:
            1. Use the scenario and user profile as the guiding context for the user's next input.
            2. Ensure the next input feels natural, conversational, and relevant to the last assistant reply in the conversation.
            3. Keep the tone consistent with the previous user inputs.
            4. The generated user input should be concise, ideally no more than 1-2 sentences.

            IMPORTANT: The output must be formatted as a JSON object with a single key `simulated_input`, 
            where the value is the generated user input in {language}.

            Example Language: english
            Example User Profile: "Jeff Seid, is available Monday and Thursday afternoons, and their phone number is 0010281839."
            Example Scenario: "A user seeking tips for securing a funding round."
            Example Previous Conversation:
            [
                {{"role": "user", "content": "Hi, I need help preparing for my funding pitch."}},
                {{"role": "assistant", "content": "Of course! Can you share more about your business and the type of investors you are targeting?"}}
            ]
            Example JSON Output:
            {{
                "simulated_input": "Sure, we are a SaaS startup focusing on productivity tools for small businesses."
            }}

            Language: {language}
            User Profile: "{golden.user_description}"
            Scenario: "{golden.scenario}"
            Previous Conversation:
            {previous_conversation}

            JSON Output:
        """
        )
        return prompt

    @staticmethod
    def stop_simulation(
        previous_conversation: str, expected_outcome: str
    ) -> str:
        prompt = textwrap.dedent(
            f"""You are a Conversation Completion Checker.
            Your task is to determine whether the conversation has achieved the expected outcome and should be terminated.

            Guidelines:
            1. Review the entire conversation and decide if the expected outcome has been met and the conversation has ended.
            2. If the expected outcome has been met, mark the conversation as complete.
            3. If not, mark it as incomplete and briefly describe what remains to be done.

            IMPORTANT: The output must be formatted as a JSON object with two keys:
            `is_complete` (a boolean) and `reason` (a string).

            Example Expected Outcome: "The user has succesfully reset their password."
            Example Conversation History:
            [
                {{"role": "user", "content": "I forgot my password and need to reset it."}},
                {{"role": "assistant", "content": "Sure. First, go to the login page and click 'Forgot Password'."}},
            ]
            Example JSON Output:
            {{
                "is_complete": False,
                "reason": "The assistant explained how to forget password but ahas not confirmed that the user successfully set a new password."
            }}

            Expected Outcome: "{expected_outcome}"
            Conversation History:
            {previous_conversation}
            JSON Output:
            """
        )
        return prompt



from ollama import Client, AsyncClient
from typing import List

from deepeval.key_handler import EmbeddingKeyValues, KEY_FILE_HANDLER
from deepeval.models import DeepEvalBaseEmbeddingModel
from deepeval.models.retry_policy import (
    create_retry_decorator,
)
from deepeval.constants import ProviderSlug as PS


retry_ollama = create_retry_decorator(PS.OLLAMA)


class OllamaEmbeddingModel(DeepEvalBaseEmbeddingModel):
    def __init__(self, *args, **kwargs):
        self.base_url = KEY_FILE_HANDLER.fetch_data(
            EmbeddingKeyValues.LOCAL_EMBEDDING_BASE_URL
        )
        model_name = KEY_FILE_HANDLER.fetch_data(
            EmbeddingKeyValues.LOCAL_EMBEDDING_MODEL_NAME
        )
        # TODO: This is not being used. Clean it up in consistency PR
        self.api_key = KEY_FILE_HANDLER.fetch_data(
            EmbeddingKeyValues.LOCAL_EMBEDDING_API_KEY
        )
        self.args = args
        self.kwargs = kwargs
        super().__init__(model_name)

    @retry_ollama
    def embed_text(self, text: str) -> List[float]:
        embedding_model = self.load_model()
        response = embedding_model.embed(
            model=self.model_name,
            input=text,
        )
        return response["embeddings"][0]

    @retry_ollama
    def embed_texts(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model()
        response = embedding_model.embed(
            model=self.model_name,
            input=texts,
        )
        return response["embeddings"]

    @retry_ollama
    async def a_embed_text(self, text: str) -> List[float]:
        embedding_model = self.load_model(async_mode=True)
        response = await embedding_model.embed(
            model=self.model_name,
            input=text,
        )
        return response["embeddings"][0]

    @retry_ollama
    async def a_embed_texts(self, texts: List[str]) -> List[List[float]]:
        embedding_model = self.load_model(async_mode=True)
        response = await embedding_model.embed(
            model=self.model_name,
            input=texts,
        )
        return response["embeddings"]

    ###############################################
    # Model
    ###############################################

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(Client)
        return self._build_client(AsyncClient)

    def _build_client(self, cls):
        return cls(host=self.base_url, **self.kwargs)

    def get_model_name(self):
        return f"{self.model_name} (Ollama)"



import json
import uuid
from typing import Any, List, Optional

from deepeval.tracing.types import ToolSpan, TraceSpanStatus
from deepeval.tracing.context import current_span_context
from deepeval.utils import shorten, len_long
from deepeval.openai.types import OutputParameters


_URL_MAX = 200
_JSON_MAX = max(
    len_long(), 400
)  # <- make this bigger by increasing DEEPEVAL_MAXLEN_LONG above 400


def _compact_dump(value: Any) -> str:
    try:
        dumped = json.dumps(
            value, ensure_ascii=False, default=str, separators=(",", ":")
        )
    except Exception:
        dumped = repr(value)
    return shorten(dumped, max_len=_JSON_MAX)


def _fmt_url(url: Optional[str]) -> str:
    if not url:
        return ""
    if url.startswith("data:"):
        return "[data-uri]"
    return shorten(url, max_len=_URL_MAX)


def create_child_tool_spans(output_parameters: OutputParameters):

    if output_parameters.tools_called is None:
        return

    current_span = current_span_context.get()
    for tool_called in output_parameters.tools_called:
        tool_span = ToolSpan(
            **{
                "uuid": str(uuid.uuid4()),
                "trace_uuid": current_span.trace_uuid,
                "parent_uuid": current_span.uuid,
                "start_time": current_span.start_time,
                "end_time": current_span.start_time,
                "status": TraceSpanStatus.SUCCESS,
                "children": [],
                "name": tool_called.name,
                "input": tool_called.input_parameters,
                "output": None,
                "metrics": None,
                "description": tool_called.description,
            }
        )
        current_span.children.append(tool_span)


def stringify_multimodal_content(content: Any) -> str:
    """
    Return a short, human-readable summary string for an OpenAI-style multimodal `content` value.

    This is used to populate span summaries, such as `InputParameters.input`. It never raises and
    never returns huge blobs.

    Notes:
    - Data URIs are redacted to "[data-uri]".
    - Output is capped via `deepeval.utils.shorten` (configurable through settings).
    - Fields that are not explicitly handled are returned as size-capped JSON dumps
    - This string is for display/summary only, not intended to be parsable.

    Args:
        content: The value of an OpenAI message `content`, may be a str or list of typed parts,
                 or any nested structure.

    Returns:
        A short, readable `str` summary.
    """
    if content is None:
        return ""
    if isinstance(content, str):
        return content
    if isinstance(content, (bytes, bytearray)):
        return f"[bytes:{len(content)}]"

    # list of parts for Chat & Responses
    if isinstance(content, list):
        parts: List[str] = []
        for part in content:
            s = stringify_multimodal_content(part)
            if s:
                parts.append(s)
        return "\n".join(parts)

    # documented dict shapes (Chat & Responses)
    if isinstance(content, dict):
        t = content.get("type")

        # Chat Completions
        if t == "text":
            return str(content.get("text", ""))
        if t == "image_url":
            image_url = content.get("image_url")
            if isinstance(image_url, str):
                url = image_url
            else:
                url = (image_url or {}).get("url") or content.get("url")
            return f"[image:{_fmt_url(url)}]"

        # Responses API variants
        if t == "input_text":
            return str(content.get("text", ""))
        if t == "input_image":
            image_url = content.get("image_url")
            if isinstance(image_url, str):
                url = image_url
            else:
                url = (image_url or {}).get("url") or content.get("url")
            return f"[image:{_fmt_url(url)}]"

        # readability for other input_* types we don't currently handle
        if t and t.startswith("input_"):
            return f"[{t}]"

    # unknown dicts and types returned as shortened JSON
    return _compact_dump(content)



from .handler import instrument_llama_index
from .agent.patched import FunctionAgent, ReActAgent, CodeActAgent


__all__ = [
    "instrument_llama_index",
    "FunctionAgent",
    "ReActAgent",
    "CodeActAgent",
]



import asyncio

from typing import Optional, Tuple, Union, Dict
from contextlib import AsyncExitStack
from pydantic import BaseModel

from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.models.llms.utils import trim_and_load_json, safe_asyncio_run
from deepeval.constants import ProviderSlug as PS

# check aiobotocore availability
try:
    from aiobotocore.session import get_session
    from botocore.config import Config

    aiobotocore_available = True
except ImportError:
    aiobotocore_available = False

# define retry policy
retry_bedrock = create_retry_decorator(PS.BEDROCK)


def _check_aiobotocore_available():
    if not aiobotocore_available:
        raise ImportError(
            "aiobotocore and botocore are required for this functionality. "
            "Install them via your package manager (e.g. pip install aiobotocore botocore)"
        )


class AmazonBedrockModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model_id: str,
        region_name: str,
        aws_access_key_id: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        input_token_cost: float = 0,
        output_token_cost: float = 0,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        _check_aiobotocore_available()
        super().__init__(model_id)

        self.model_id = model_id
        self.region_name = region_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.input_token_cost = input_token_cost
        self.output_token_cost = output_token_cost

        # prepare aiobotocore session, config, and async exit stack
        self._session = get_session()
        self._exit_stack = AsyncExitStack()
        self.kwargs = kwargs
        self.generation_kwargs = generation_kwargs or {}
        self._client = None
        self._sdk_retry_mode: Optional[bool] = None

    ###############################################
    # Generate functions
    ###############################################

    def generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, Dict], float]:
        return safe_asyncio_run(self.a_generate(prompt, schema))

    @retry_bedrock
    async def a_generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, Dict], float]:
        payload = self.get_converse_request_body(prompt)
        client = await self._ensure_client()
        response = await client.converse(
            modelId=self.model_id,
            messages=payload["messages"],
            inferenceConfig=payload["inferenceConfig"],
        )
        message = response["output"]["message"]["content"][0]["text"]
        cost = self.calculate_cost(
            response["usage"]["inputTokens"],
            response["usage"]["outputTokens"],
        )
        if schema is None:
            return message, cost
        else:
            json_output = trim_and_load_json(message)
            return schema.model_validate(json_output), cost

    ###############################################
    # Client management
    ###############################################

    async def _ensure_client(self):
        use_sdk = sdk_retries_for(PS.BEDROCK)

        # only rebuild if client is missing or the sdk retry mode changes
        if self._client is None or self._sdk_retry_mode != use_sdk:
            # Close any previous
            if self._client is not None:
                await self._exit_stack.aclose()
                self._client = None

            # create retry config for botocore
            retries_config = {"max_attempts": (5 if use_sdk else 1)}
            if use_sdk:
                retries_config["mode"] = "adaptive"

            config = Config(retries=retries_config)

            cm = self._session.create_client(
                "bedrock-runtime",
                region_name=self.region_name,
                aws_access_key_id=self.aws_access_key_id,
                aws_secret_access_key=self.aws_secret_access_key,
                config=config,
                **self.kwargs,
            )
            self._client = await self._exit_stack.enter_async_context(cm)
            self._sdk_retry_mode = use_sdk

        return self._client

    async def close(self):
        await self._exit_stack.aclose()
        self._client = None

    ###############################################
    # Helpers
    ###############################################

    def get_converse_request_body(self, prompt: str) -> dict:

        return {
            "messages": [{"role": "user", "content": [{"text": prompt}]}],
            "inferenceConfig": {
                **self.generation_kwargs,
            },
        }

    def calculate_cost(self, input_tokens: int, output_tokens: int) -> float:
        return (
            input_tokens * self.input_token_cost
            + output_tokens * self.output_token_cost
        )

    def load_model(self):
        pass

    def get_model_name(self) -> str:
        return self.model_id



from .llm_test_case import (
    LLMTestCase,
    LLMTestCaseParams,
    ToolCall,
    ToolCallParams,
)
from .conversational_test_case import (
    ConversationalTestCase,
    Turn,
    TurnParams,
)
from .mllm_test_case import MLLMTestCase, MLLMTestCaseParams, MLLMImage
from .arena_test_case import ArenaTestCase
from .mcp import (
    MCPServer,
    MCPPromptCall,
    MCPResourceCall,
    MCPToolCall,
)


__all__ = [
    "LLMTestCase",
    "LLMTestCaseParams",
    "ToolCall",
    "ToolCallParams",
    "ConversationalTestCase",
    "Turn",
    "TurnParams",
    "MCPServer",
    "MCPPromptCall",
    "MCPResourceCall",
    "MCPToolCall",
    "MLLMTestCase",
    "MLLMTestCaseParams",
    "MLLMImage",
    "ArenaTestCase",
]



from typing import List, Optional, Dict, Union
from tqdm import tqdm

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import (
    DeepEvalBaseBenchmark,
    DeepEvalBaseBenchmarkResult,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.gsm8k.template import GSM8KTemplate
from deepeval.benchmarks.schema import NumberSchema
from deepeval.telemetry import capture_benchmark_run


class GSM8K(DeepEvalBaseBenchmark):
    def __init__(
        self,
        n_shots: int = 3,
        enable_cot: bool = True,
        n_problems: int = 1319,
        verbose_mode: bool = False,
        confinement_instructions: Optional[str] = None,
        **kwargs,
    ):
        from deepeval.scorer import Scorer
        import pandas as pd

        assert n_shots <= 15, "GSM8K only supports n_shots <= 15"
        super().__init__(**kwargs)
        self.scorer = Scorer()
        self.shots_dataset: List[Dict] = None
        self.n_shots: int = n_shots
        self.enable_cot: bool = enable_cot
        self.n_problems: int = n_problems
        self.predictions: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None
        self.verbose_mode = verbose_mode
        if not confinement_instructions:
            self.confinement_instructions = (
                "Make sure to output only the numerical answer."
            )
        else:
            self.confinement_instructions = confinement_instructions

    def evaluate(
        self, model: DeepEvalBaseLLM, *args, **kwargs
    ) -> DeepEvalBaseBenchmarkResult:
        import pandas as pd

        with capture_benchmark_run("GSM8K", len(self.tasks)):
            overall_correct_predictions = 0
            overall_total_predictions = self.n_problems
            predictions_row = []

            # Solving each problem
            goldens = self.load_benchmark_dataset()[: self.n_problems]
            for idx, golden in enumerate(
                tqdm(goldens, desc=f"Processing {self.n_problems} problems")
            ):
                result = self.predict(model, golden)
                prediction = result["prediction"]
                score = result["score"]

                if score:
                    overall_correct_predictions += 1
                predictions_row.append(
                    (golden.input, prediction, golden.expected_output, score)
                )
                if self.verbose_mode:
                    self.print_verbose_logs(
                        idx,
                        golden.input,
                        golden.expected_output,
                        prediction,
                        score,
                    )

            # Calculate overall accuracy
            overall_accuracy = (
                overall_correct_predictions / overall_total_predictions
            )
            print(f"Overall GSM8K Accuracy: {overall_accuracy}")

            self.predictions = pd.DataFrame(
                predictions_row,
                columns=["Input", "Prediction", "Expected Output", "Correct"],
            )
            self.overall_score = overall_accuracy

            return DeepEvalBaseBenchmarkResult(
                overall_accuracy=overall_accuracy
            )

    def predict(self, model: DeepEvalBaseLLM, golden: Golden) -> Dict:
        # Define prompt template
        assert (
            self.shots_dataset != None
        ), "Example dataset is empty. Call load_benchmark."
        prompt: dict = GSM8KTemplate.generate_output(
            train_set=self.shots_dataset,
            input=golden.input,
            n_shots=self.n_shots,
            enable_cot=self.enable_cot,
        )

        # Enforced model generation
        prediction = None
        try:
            res: NumberSchema = model.generate(
                prompt=prompt, schema=NumberSchema
            )
            prediction = self._extract_prediction_from_response(res)
        except (TypeError, AttributeError) as e:

            prompt += f"\n\n{self.confinement_instructions}"
            res = model.generate(prompt)
            prediction = self._extract_prediction_from_response(res)

        # For native models, shouldn't happen but just in case
        if isinstance(prediction, tuple):
            prediction = prediction[0]
        prediction = str(prediction)

        score = self.scorer.exact_match_score(
            golden.expected_output, prediction
        )

        return {"prediction": prediction, "score": score}

    def _extract_prediction_from_response(self, res) -> str:
        """
        Extract prediction from model response, handling various response types.
        """
        # Case 1: Response has .answer attribute (NumberSchema case)
        if hasattr(res, "answer"):
            return str(res.answer)

        # Case 2: Response is a tuple
        elif isinstance(res, tuple):
            return self._extract_from_tuple(res)

        else:
            return str(res)

    def _extract_from_tuple(self, res: tuple) -> str:
        """Extract prediction from tuple response."""
        if len(res) == 0:
            return ""
        first_elem = res[0]
        if hasattr(first_elem, "answer"):
            return str(first_elem.answer)

    def load_benchmark_dataset(self) -> List[Golden]:
        from datasets import load_dataset

        # Load dataset
        if self.dataset:
            dataset = self.dataset
        else:
            dataset = load_dataset("gsm8k", "main")
            self.dataset = dataset

        # Construct example dataset for n_shot inference
        if not self.shots_dataset:
            train_set = dataset["train"]
            shots_set = []
            for data in train_set:
                shots_set.append(data)
            self.shots_dataset = shots_set

        # Construct test set
        goldens: List[Golden] = []
        for data in dataset["test"]:
            input = data["question"]
            output = GSM8KTemplate.format_answer(data)
            golden = Golden(input=input, expected_output=output)
            goldens.append(golden)

        return goldens

    def print_verbose_logs(
        self,
        idx: int,
        input: str,
        expected_output: str,
        prediction: str,
        score: int,
    ) -> str:
        steps = [
            f"Input:\n{input}",
            f"Score: {score}\nPrediction: {prediction}\nExpected Output: {expected_output}",
        ]
        verbose_logs = ""
        for i in range(len(steps) - 1):
            verbose_logs += steps[i]

            # don't add new line for penultimate step
            if i < len(steps) - 2:
                verbose_logs += " \n \n"

        if self.verbose_mode:
            print("*" * 50)
            print(f"Problem {idx + 1}")
            print("*" * 50)
            print("")
            print(verbose_logs + f"\n \n{steps[-1]}")
            print("")
            print("=" * 70)

        return verbose_logs



from typing import List
from deepeval.test_case import ToolCall
import textwrap


class ArgumentCorrectnessTemplate:
    @staticmethod
    def generate_verdicts(input: str, tools_called: List[ToolCall]):

        stringified_tools_called = repr(tools_called)

        return textwrap.dedent(
            f"""
            For the provided list of tool calls, determine whether each tool call input parameter is relevantly and correctly addresses the input.

            Please generate a list of JSON with two keys: `verdict` and `reason`.
            The 'verdict' key should STRICTLY be either a 'yes' or 'no'. Answer 'yes' if the tool call input parameter is relevantly and correctly addresses the original input, 'no' if the tool call input parameter doesn't correctly and relevantly address the original input.
            The 'reason' is the reason for the verdict.
            Provide a 'reason' ONLY if the answer is 'no'. 
            If there is no input parameter, answer 'no' for the verdict and provide the reason as "No input parameter provided".

            **
            IMPORTANT: Please make sure to only return in valid and parseable JSON format, with the 'verdicts' key mapping to a list of JSON objects. Ensure all strings are closed appropriately. Repair any invalid JSON before you output it.
            Example input: 
            "What was the highest temperature recorded in Paris in 2023?"
            
            Example tool calls: 
            [
                ToolCall(
                    name="WeatherHistoryAPI",
                    description="Fetches historical weather data for a given city and date range",
                    reasoning="I need to check all 2023 temperature records for Paris to find the highest one.",
                    input_parameters={{
                        "city_name": "Paris",
                        "country_code": "FR",
                        "date_range_start": "2023-01-01",
                        "date_range_end": "2023-12-31",
                        "data_type": "temperature_max_daily_celsius"
                    }}
                ),
                ToolCall(
                    name="MathAnalyzer",
                    description="Performs statistical calculations on numeric datasets",
                    reasoning="I will calculate the maximum temperature value from the daily dataset.",
                    input_parameters={{
                        "operation": "max",
                        "dataset_source": "WeatherHistoryAPI.daily_max_temperatures",
                        "expected_unit": "celsius"
                    }}
                ),
                ToolCall(
                    name="MovieRecommender",
                    description="Recommends movies based on user mood or location",
                    reasoning="I thought Paris movies might be fun to suggest, but this is unrelated to the question.",
                    input_parameters={{
                        "preferred_genres": ["romance", "comedy"],
                        "setting_city": "Paris",
                        "language_preference": "French or English"
                    }}
                )
            ]

            Example JSON:
            {{
                "verdicts": [
                    {{
                        "verdict": "yes"
                    }},
                    {{
                        "verdict": "yes"
                    }},
                    {{
                        "verdict": "no",
                        "reason": "Recommending romantic Parisian comedies does not help find the highest temperature in 2023."
                    }}
                ]  
            }}
            ===== END OF EXAMPLE ======

            Since you are going to generate a verdict for each statement, the number of 'verdicts' SHOULD BE STRICTLY EQUAL to the number of `statements`.
            **          

            Input:
            {input}

            Tool Calls:
            {stringified_tools_called}

            JSON:
            """
        )

    @staticmethod
    def generate_reason(
        incorrect_tool_calls_reasons: List[str], input: str, score: float
    ):
        return textwrap.dedent(
            f"""Given the argument correctness score, the list of reasons of incorrect tool calls, and the input, provide a CONCISE reason for the score. Explain why it is not higher, but also why it is at its current score. You can mention tool calls or input, but do not mention an output or a response.
            If there is nothing incorrect, just say something positive with an upbeat encouraging tone (but don't overdo it otherwise it gets annoying).

            **
            IMPORTANT: Please make sure to only return in JSON format, with the 'reason' key providing the reason. Ensure all strings are closed appropriately. Repair any invalid JSON before you output it.

            Example:
            Example JSON:
            {{
                "reason": "The score is <argument_correctness_score> because <your_reason>."
            }}
            ===== END OF EXAMPLE ======
            **


            Argument Correctness Score:
            {score}

            Reasons why the score can't be higher based on incorrect tool calls:
            {incorrect_tool_calls_reasons}

            Input:
            {input}

            JSON:
             """
        )



import re
import uuid
from jinja2 import Template
from typing import Any, Dict, Type, Optional, List
from pydantic import BaseModel, create_model

from deepeval.prompt.api import (
    PromptInterpolationType,
    OutputSchema,
    SchemaDataType,
    OutputSchemaField,
)

###################################
# Interpolation
###################################


def interpolate_mustache(text: str, **kwargs) -> str:
    """Interpolate using Mustache format: {{variable}}"""
    formatted_template = re.sub(r"\{\{(\w+)\}\}", r"{\1}", text)
    return formatted_template.format(**kwargs)


def interpolate_mustache_with_space(text: str, **kwargs) -> str:
    """Interpolate using Mustache with space format: {{ variable }}"""
    formatted_template = re.sub(r"\{\{ (\w+) \}\}", r"{\1}", text)
    return formatted_template.format(**kwargs)


def interpolate_fstring(text: str, **kwargs) -> str:
    """Interpolate using F-string format: {variable}"""
    return text.format(**kwargs)


def interpolate_dollar_brackets(text: str, **kwargs) -> str:
    """Interpolate using Dollar Brackets format: ${variable}"""
    formatted_template = re.sub(r"\$\{(\w+)\}", r"{\1}", text)
    return formatted_template.format(**kwargs)


def interpolate_jinja(text: str, **kwargs) -> str:
    template = Template(text)
    return template.render(**kwargs)


def interpolate_text(
    interpolation_type: PromptInterpolationType, text: str, **kwargs
) -> str:
    """Apply the appropriate interpolation method based on the type"""
    if interpolation_type == PromptInterpolationType.MUSTACHE:
        return interpolate_mustache(text, **kwargs)
    elif interpolation_type == PromptInterpolationType.MUSTACHE_WITH_SPACE:
        return interpolate_mustache_with_space(text, **kwargs)
    elif interpolation_type == PromptInterpolationType.FSTRING:
        return interpolate_fstring(text, **kwargs)
    elif interpolation_type == PromptInterpolationType.DOLLAR_BRACKETS:
        return interpolate_dollar_brackets(text, **kwargs)
    elif interpolation_type == PromptInterpolationType.JINJA:
        return interpolate_jinja(text, **kwargs)


###################################
# Output Schema Deconstruction
###################################

schema_type_map: Dict[str, Any] = {
    SchemaDataType.STRING.value: str,
    SchemaDataType.INTEGER.value: int,
    SchemaDataType.FLOAT.value: float,
    SchemaDataType.BOOLEAN.value: bool,
    SchemaDataType.NULL.value: type(None),
    SchemaDataType.OBJECT.value: dict,
}


def construct_nested_base_model(
    parent: OutputSchemaField,
    parent_id_map: Dict[Optional[str], List[OutputSchemaField]],
    model_name: str,
) -> Type[BaseModel]:
    child_fields: Dict[str, tuple] = {}
    for child in parent_id_map.get(parent.id, []):
        child_type = (
            child.type.value if hasattr(child.type, "value") else child.type
        )
        if child_type == SchemaDataType.OBJECT.value:
            python_type = construct_nested_base_model(
                child, parent_id_map, child.name
            )
        else:
            python_type = schema_type_map.get(child_type, Any)
        default = ... if child.required else None
        child_fields[child.name or child.id] = (python_type, default)
    return create_model(model_name, **child_fields)


def construct_base_model(
    schema: Optional[OutputSchema] = None,
) -> Type[BaseModel]:
    if not schema:
        return None
    if not schema.fields:
        return create_model(schema.name)

    parent_id_map: Dict[Optional[str], List[OutputSchemaField]] = {}
    for field in schema.fields:
        parent_id = field.parent_id or None
        if parent_id_map.get(parent_id) is None:
            parent_id_map[parent_id] = []
        parent_id_map[parent_id].append(field)

    root_fields: Dict[str, tuple] = {}
    for field in parent_id_map.get(None, []):
        field_type = (
            field.type.value if hasattr(field.type, "value") else field.type
        )
        if field_type == SchemaDataType.OBJECT.value:
            python_type = construct_nested_base_model(
                field, parent_id_map, field.name
            )
        else:
            python_type = schema_type_map.get(field_type, Any)
        default = ... if field.required else None
        root_fields[field.name] = (python_type, default)

    return create_model(schema.name, **root_fields)


###################################
# Output Schema Construction
###################################


def _process_model(
    model_class: Type[BaseModel],
    parent_id: Optional[str] = None,
) -> List[OutputSchemaField]:
    fields = []
    model_fields = model_class.model_fields
    for field_name, field_info in model_fields.items():
        field_id = str(uuid.uuid4())
        annotation = field_info.annotation
        field_type = "STRING"
        if annotation == str:
            field_type = "STRING"
        elif annotation == int:
            field_type = "INTEGER"
        elif annotation == float:
            field_type = "FLOAT"
        elif annotation == bool:
            field_type = "BOOLEAN"
        elif annotation == list:
            raise ValueError("Unsupported structured output: list")
        elif annotation == dict:
            raise ValueError("Unsupported structured output: dict")
        elif (
            hasattr(annotation, "__bases__")
            and BaseModel in annotation.__bases__
        ):
            field_type = "OBJECT"
            parent_field = OutputSchemaField(
                id=field_id,
                name=field_name,
                type=field_type,
                required=field_info.default is ...,
                parent_id=parent_id,
            )
            fields.append(parent_field)
            nested_fields = _process_model(annotation, field_id)
            fields.extend(nested_fields)
            continue
        required = field_info.default is ...
        fields.append(
            OutputSchemaField(
                id=field_id,
                name=field_name,
                type=field_type,
                required=required,
                parent_id=parent_id,
            )
        )
    return fields


def construct_output_schema(
    base_model_class: Optional[Type[BaseModel]] = None,
) -> Optional[OutputSchema]:
    if base_model_class is None:
        return None
    all_fields = _process_model(base_model_class)
    return OutputSchema(fields=all_fields, name=base_model_class.__name__)



from typing import List
from pydantic import BaseModel


class ContextualRecallVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[ContextualRecallVerdict]


class ContextualRecallScoreReason(BaseModel):
    reason: str



from typing import Optional, Tuple, Union, Dict
from pydantic import BaseModel
from openai import OpenAI, AsyncOpenAI
from openai.types.chat import ChatCompletion

from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.models.llms.utils import trim_and_load_json
from deepeval.models import DeepEvalBaseLLM
from deepeval.key_handler import ModelKeyValues, KEY_FILE_HANDLER
from deepeval.constants import ProviderSlug as PS


# consistent retry rules
retry_local = create_retry_decorator(PS.LOCAL)


class LocalModel(DeepEvalBaseLLM):
    def __init__(
        self,
        model: Optional[str] = None,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        temperature: float = 0,
        format: Optional[str] = None,
        generation_kwargs: Optional[Dict] = None,
        **kwargs,
    ):
        model_name = model or KEY_FILE_HANDLER.fetch_data(
            ModelKeyValues.LOCAL_MODEL_NAME
        )
        self.local_model_api_key = api_key or KEY_FILE_HANDLER.fetch_data(
            ModelKeyValues.LOCAL_MODEL_API_KEY
        )
        self.base_url = base_url or KEY_FILE_HANDLER.fetch_data(
            ModelKeyValues.LOCAL_MODEL_BASE_URL
        )
        self.format = format or KEY_FILE_HANDLER.fetch_data(
            ModelKeyValues.LOCAL_MODEL_FORMAT
        )
        if temperature < 0:
            raise ValueError("Temperature must be >= 0.")
        self.temperature = temperature
        self.kwargs = kwargs
        self.generation_kwargs = generation_kwargs or {}
        super().__init__(model_name)

    ###############################################
    # Other generate functions
    ###############################################

    @retry_local
    def generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, Dict], float]:
        client = self.load_model(async_mode=False)
        response: ChatCompletion = client.chat.completions.create(
            model=self.model_name,
            messages=[{"role": "user", "content": prompt}],
            temperature=self.temperature,
            **self.generation_kwargs,
        )
        res_content = response.choices[0].message.content

        if schema:
            json_output = trim_and_load_json(res_content)
            return schema.model_validate(json_output), 0.0
        else:
            return res_content, 0.0

    @retry_local
    async def a_generate(
        self, prompt: str, schema: Optional[BaseModel] = None
    ) -> Tuple[Union[str, Dict], float]:
        client = self.load_model(async_mode=True)
        response: ChatCompletion = await client.chat.completions.create(
            model=self.model_name,
            messages=[{"role": "user", "content": prompt}],
            temperature=self.temperature,
            **self.generation_kwargs,
        )
        res_content = response.choices[0].message.content

        if schema:
            json_output = trim_and_load_json(res_content)
            return schema.model_validate(json_output), 0.0
        else:
            return res_content, 0.0

    ###############################################
    # Model
    ###############################################

    def get_model_name(self):
        model_name = KEY_FILE_HANDLER.fetch_data(
            ModelKeyValues.LOCAL_MODEL_NAME
        )
        return f"{model_name} (Local Model)"

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(OpenAI)
        return self._build_client(AsyncOpenAI)

    def _client_kwargs(self) -> Dict:
        """
        If Tenacity manages retries, turn off OpenAI SDK retries to avoid double retrying.
        If users opt into SDK retries via DEEPEVAL_SDK_RETRY_PROVIDERS=local, leave them enabled.
        """
        kwargs = dict(self.kwargs or {})
        if not sdk_retries_for(PS.LOCAL):
            kwargs["max_retries"] = 0
        return kwargs

    def _build_client(self, cls):
        kw = dict(
            api_key=self.local_model_api_key,
            base_url=self.base_url,
            **self._client_kwargs(),
        )
        try:
            return cls(**kw)
        except TypeError as e:
            # Older OpenAI SDKs may not accept max_retries; drop and retry once.
            if "max_retries" in str(e):
                kw.pop("max_retries", None)
                return cls(**kw)
            raise



from typing import Set, Dict, Optional, Union
import inspect

from deepeval.metrics.dag import (
    BaseNode,
    BinaryJudgementNode,
    NonBinaryJudgementNode,
    VerdictNode,
    TaskNode,
    DeepAcyclicGraph,
)
from deepeval.metrics.conversational_dag import (
    ConversationalBaseNode,
    ConversationalBinaryJudgementNode,
    ConversationalNonBinaryJudgementNode,
    ConversationalTaskNode,
    ConversationalVerdictNode,
)
from deepeval.test_case import LLMTestCaseParams, TurnParams


def is_valid_dag_from_roots(
    root_nodes: Union[list[BaseNode], list[ConversationalBaseNode]],
    multiturn: bool,
) -> bool:
    visited = set()
    for root in root_nodes:
        if not is_valid_dag(root, multiturn, visited, set()):
            return False
    return True


def is_valid_dag(
    node: Union[BaseNode, ConversationalBaseNode],
    multiturn: bool,
    visited=None,
    stack=None,
) -> bool:
    if visited is None:
        visited = set()
    if stack is None:
        stack = set()

    if node in stack:
        return False
    if node in visited:
        return True

    visited.add(node)
    stack.add(node)
    if not multiturn:
        if (
            isinstance(node, TaskNode)
            or isinstance(node, BinaryJudgementNode)
            or isinstance(node, NonBinaryJudgementNode)
        ):
            for child in node.children:
                if not is_valid_dag(child, multiturn, visited, stack):
                    return False
    else:
        if (
            isinstance(node, ConversationalTaskNode)
            or isinstance(node, ConversationalBinaryJudgementNode)
            or isinstance(node, ConversationalNonBinaryJudgementNode)
        ):
            for child in node.children:
                if not is_valid_dag(child, multiturn, visited, stack):
                    return False

    stack.remove(node)
    return True


def extract_required_params(
    nodes: list[BaseNode],
    multiturn: bool,
    required_params: Optional[
        Union[Set[LLMTestCaseParams], Set[TurnParams]]
    ] = None,
) -> Union[Set[LLMTestCaseParams], Set[TurnParams]]:
    if required_params is None:
        required_params = set()

    for node in nodes:
        if not multiturn:
            if (
                isinstance(node, TaskNode)
                or isinstance(node, BinaryJudgementNode)
                or isinstance(node, NonBinaryJudgementNode)
            ):
                if node.evaluation_params is not None:
                    required_params.update(node.evaluation_params)
                extract_required_params(
                    node.children, multiturn, required_params
                )
        else:
            if (
                isinstance(node, ConversationalTaskNode)
                or isinstance(node, ConversationalBinaryJudgementNode)
                or isinstance(node, ConversationalNonBinaryJudgementNode)
            ):
                if node.evaluation_params is not None:
                    required_params.update(node.evaluation_params)
                extract_required_params(
                    node.children, multiturn, required_params
                )

    return required_params


def copy_graph(original_dag: DeepAcyclicGraph) -> DeepAcyclicGraph:
    # This mapping avoids re-copying nodes that appear in multiple places.
    visited: Union[
        Dict[BaseNode, BaseNode],
        Dict[ConversationalBaseNode, ConversationalBaseNode],
    ] = {}

    def copy_node(
        node: Union[BaseNode, ConversationalBaseNode],
    ) -> Union[BaseNode, ConversationalBaseNode]:
        if node in visited:
            return visited[node]

        node_class = type(node)
        args = vars(node)
        superclasses = node_class.__mro__
        valid_params = []
        for superclass in superclasses:
            signature = inspect.signature(superclass.__init__)
            superclass_params = signature.parameters.keys()
            valid_params.extend(superclass_params)
        valid_params = set(valid_params)
        valid_args = {
            key: args[key]
            for key in valid_params
            if key in args
            and key
            not in [
                "children",
                "child",
                "_parents",
                "_parent",
                "_verdict",
                "_indegree",
                "_depth",
            ]
        }
        if not original_dag.multiturn:
            if (
                isinstance(node, TaskNode)
                or isinstance(node, BinaryJudgementNode)
                or isinstance(node, NonBinaryJudgementNode)
            ):
                copied_node = node_class(
                    **valid_args,
                    children=[copy_node(child) for child in node.children]
                )
            else:
                if isinstance(node, VerdictNode) and node.child:
                    copied_node = node_class(
                        **valid_args, child=copy_node(node.child)
                    )
                else:
                    copied_node = node_class(**valid_args)
        else:
            if (
                isinstance(node, ConversationalTaskNode)
                or isinstance(node, ConversationalBinaryJudgementNode)
                or isinstance(node, ConversationalNonBinaryJudgementNode)
            ):
                copied_node = node_class(
                    **valid_args,
                    children=[copy_node(child) for child in node.children]
                )
            else:
                if isinstance(node, ConversationalVerdictNode) and node.child:
                    copied_node = node_class(
                        **valid_args, child=copy_node(node.child)
                    )
                else:
                    copied_node = node_class(**valid_args)

        visited[node] = copied_node
        return copied_node

    # Copy all root nodes (the recursion handles the rest).
    new_root_nodes = [copy_node(root) for root in original_dag.root_nodes]
    return DeepAcyclicGraph(new_root_nodes)



from pydantic import (
    ConfigDict,
    Field,
    BaseModel,
    model_validator,
    PrivateAttr,
    AliasChoices,
)
from typing import List, Optional, Dict, Any
from enum import Enum
import json
import uuid

from deepeval.test_case.mcp import (
    MCPServer,
    MCPPromptCall,
    MCPResourceCall,
    MCPToolCall,
    validate_mcp_servers,
)


class LLMTestCaseParams(Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"
    RETRIEVAL_CONTEXT = "retrieval_context"
    TOOLS_CALLED = "tools_called"
    EXPECTED_TOOLS = "expected_tools"
    MCP_SERVERS = "mcp_servers"
    MCP_TOOLS_CALLED = "mcp_tools_called"
    MCP_RESOURCES_CALLED = "mcp_resources_called"
    MCP_PROMPTS_CALLED = "mcp_prompts_called"


class ToolCallParams(Enum):
    INPUT_PARAMETERS = "input_parameters"
    OUTPUT = "output"


def _make_hashable(obj):
    """
    Convert an object to a hashable representation recursively.

    Args:
        obj: The object to make hashable

    Returns:
        A hashable representation of the object
    """
    if obj is None:
        return None
    elif isinstance(obj, dict):
        # Convert dict to tuple of sorted key-value pairs
        return tuple(sorted((k, _make_hashable(v)) for k, v in obj.items()))
    elif isinstance(obj, (list, tuple)):
        # Convert list/tuple to tuple of hashable elements
        return tuple(_make_hashable(item) for item in obj)
    elif isinstance(obj, set):
        # Convert set to frozenset of hashable elements
        return frozenset(_make_hashable(item) for item in obj)
    elif isinstance(obj, frozenset):
        # Handle frozenset that might contain unhashable elements
        return frozenset(_make_hashable(item) for item in obj)
    else:
        # For primitive hashable types (str, int, float, bool, etc.)
        return obj


class ToolCall(BaseModel):
    name: str
    description: Optional[str] = None
    reasoning: Optional[str] = None
    output: Optional[Any] = None
    input_parameters: Optional[Dict[str, Any]] = Field(
        None,
        serialization_alias="inputParameters",
        validation_alias=AliasChoices("inputParameters", "input_parameters"),
    )

    def __eq__(self, other):
        if not isinstance(other, ToolCall):
            return False
        return (
            self.name == other.name
            and self.input_parameters == other.input_parameters
            and self.output == other.output
        )

    def __hash__(self):
        """
        Generate a hash for the ToolCall instance.

        This method handles complex input parameters and outputs that may contain
        unhashable types like lists, dicts, and nested structures.

        Returns:
            int: Hash value for this ToolCall instance
        """
        # Handle input_parameters
        input_params = (
            self.input_parameters if self.input_parameters is not None else {}
        )
        input_params_hashable = _make_hashable(input_params)

        # Handle output - use the new helper function instead of manual handling
        output_hashable = _make_hashable(self.output)

        return hash((self.name, input_params_hashable, output_hashable))

    def __repr__(self):
        fields = []

        # Add basic fields
        if self.name:
            fields.append(f'name="{self.name}"')
        if self.description:
            fields.append(f'description="{self.description}"')
        if self.reasoning:
            fields.append(f'reasoning="{self.reasoning}"')

        # Handle nested fields like input_parameters
        if self.input_parameters:
            formatted_input = json.dumps(self.input_parameters, indent=4)
            formatted_input = self._indent_nested_field(
                "input_parameters", formatted_input
            )
            fields.append(formatted_input)

        # Handle nested fields like output
        if isinstance(self.output, dict):
            formatted_output = json.dumps(self.output, indent=4)
            formatted_output = self._indent_nested_field(
                "output", formatted_output
            )
            fields.append(formatted_output)
        elif self.output is not None:
            fields.append(f"output={repr(self.output)}")

        # Combine fields with proper formatting
        fields_str = ",\n    ".join(fields)
        return f"ToolCall(\n    {fields_str}\n)"

    @staticmethod
    def _indent_nested_field(field_name: str, formatted_field: str) -> str:
        """Helper method to indent multi-line fields for better readability."""
        lines = formatted_field.splitlines()
        return f"{field_name}={lines[0]}\n" + "\n".join(
            f"    {line}" for line in lines[1:]
        )


class LLMTestCase(BaseModel):
    model_config = ConfigDict(extra="ignore")

    input: str
    actual_output: Optional[str] = Field(
        default=None,
        serialization_alias="actualOutput",
        validation_alias=AliasChoices("actualOutput", "actual_output"),
    )
    expected_output: Optional[str] = Field(
        default=None,
        serialization_alias="expectedOutput",
        validation_alias=AliasChoices("expectedOutput", "expected_output"),
    )
    context: Optional[List[str]] = Field(
        default=None, serialization_alias="context"
    )
    retrieval_context: Optional[List[str]] = Field(
        default=None,
        serialization_alias="retrievalContext",
        validation_alias=AliasChoices("retrievalContext", "retrieval_context"),
    )
    additional_metadata: Optional[Dict] = Field(
        default=None,
        serialization_alias="additionalMetadata",
        validation_alias=AliasChoices(
            "additionalMetadata", "additional_metadata"
        ),
    )
    tools_called: Optional[List[ToolCall]] = Field(
        default=None,
        serialization_alias="toolsCalled",
        validation_alias=AliasChoices("toolsCalled", "tools_called"),
    )
    comments: Optional[str] = Field(
        default=None, serialization_alias="comments"
    )
    expected_tools: Optional[List[ToolCall]] = Field(
        default=None,
        serialization_alias="expectedTools",
        validation_alias=AliasChoices("expectedTools", "expected_tools"),
    )
    token_cost: Optional[float] = Field(
        default=None,
        serialization_alias="tokenCost",
        validation_alias=AliasChoices("tokenCost", "token_cost"),
    )
    completion_time: Optional[float] = Field(
        default=None,
        serialization_alias="completionTime",
        validation_alias=AliasChoices("completionTime", "completion_time"),
    )
    name: Optional[str] = Field(default=None)
    tags: Optional[List[str]] = Field(default=None)
    mcp_servers: Optional[List[MCPServer]] = Field(default=None)
    mcp_tools_called: Optional[List[MCPToolCall]] = Field(
        default=None,
        serialization_alias="mcpToolsCalled",
    )
    mcp_resources_called: Optional[List[MCPResourceCall]] = Field(
        default=None, serialization_alias="mcpResourcesCalled"
    )
    mcp_prompts_called: Optional[List[MCPPromptCall]] = Field(
        default=None, serialization_alias="mcpPromptsCalled"
    )
    _trace_dict: Optional[Dict] = PrivateAttr(default=None)
    _dataset_rank: Optional[int] = PrivateAttr(default=None)
    _dataset_alias: Optional[str] = PrivateAttr(default=None)
    _dataset_id: Optional[str] = PrivateAttr(default=None)
    _identifier: Optional[str] = PrivateAttr(
        default_factory=lambda: str(uuid.uuid4())
    )

    @model_validator(mode="before")
    def validate_input(cls, data):
        input = data.get("input")
        actual_output = data.get("actual_output")
        context = data.get("context")
        retrieval_context = data.get("retrieval_context")
        tools_called = data.get("tools_called")
        expected_tools = data.get("expected_tools")
        mcp_servers = data.get("mcp_servers")
        mcp_tools_called = data.get("mcp_tools_called")
        mcp_resources_called = data.get("mcp_resources_called")
        mcp_prompts_called = data.get("mcp_prompts_called")

        if input is not None:
            if not isinstance(input, str):
                raise TypeError("'input' must be a string")

        if actual_output is not None:
            if not isinstance(actual_output, str):
                raise TypeError("'actual_output' must be a string")

        # Ensure `context` is None or a list of strings
        if context is not None:
            if not isinstance(context, list) or not all(
                isinstance(item, str) for item in context
            ):
                raise TypeError("'context' must be None or a list of strings")

        # Ensure `retrieval_context` is None or a list of strings
        if retrieval_context is not None:
            if not isinstance(retrieval_context, list) or not all(
                isinstance(item, str) for item in retrieval_context
            ):
                raise TypeError(
                    "'retrieval_context' must be None or a list of strings"
                )

        # Ensure `tools_called` is None or a list of strings
        if tools_called is not None:
            if not isinstance(tools_called, list) or not all(
                isinstance(item, ToolCall) for item in tools_called
            ):
                raise TypeError(
                    "'tools_called' must be None or a list of `ToolCall`"
                )

        # Ensure `expected_tools` is None or a list of strings
        if expected_tools is not None:
            if not isinstance(expected_tools, list) or not all(
                isinstance(item, ToolCall) for item in expected_tools
            ):
                raise TypeError(
                    "'expected_tools' must be None or a list of `ToolCall`"
                )

        # Ensure `mcp_server` is None or a list of `MCPServer`
        if mcp_servers is not None:
            if not isinstance(mcp_servers, list) or not all(
                isinstance(item, MCPServer) for item in mcp_servers
            ):
                raise TypeError(
                    "'mcp_server' must be None or a list of 'MCPServer'"
                )
            else:
                validate_mcp_servers(mcp_servers)

        # Ensure `mcp_tools_called` is None or a list of `MCPToolCall`
        if mcp_tools_called is not None:
            from mcp.types import CallToolResult

            if not isinstance(mcp_tools_called, list) or not all(
                isinstance(tool_called, MCPToolCall)
                and isinstance(tool_called.result, CallToolResult)
                for tool_called in mcp_tools_called
            ):
                raise TypeError(
                    "The 'tools_called' must be a list of 'MCPToolCall' with result of type 'CallToolResult' from mcp.types"
                )

        # Ensure `mcp_resources_called` is None or a list of `MCPResourceCall`
        if mcp_resources_called is not None:
            from mcp.types import ReadResourceResult

            if not isinstance(mcp_resources_called, list) or not all(
                isinstance(resource_called, MCPResourceCall)
                and isinstance(resource_called.result, ReadResourceResult)
                for resource_called in mcp_resources_called
            ):
                raise TypeError(
                    "The 'resources_called' must be a list of 'MCPResourceCall' with result of type 'ReadResourceResult' from mcp.types"
                )

        # Ensure `mcp_prompts_called` is None or a list of `MCPPromptCall`
        if mcp_prompts_called is not None:
            from mcp.types import GetPromptResult

            if not isinstance(mcp_prompts_called, list) or not all(
                isinstance(prompt_called, MCPPromptCall)
                and isinstance(prompt_called.result, GetPromptResult)
                for prompt_called in mcp_prompts_called
            ):
                raise TypeError(
                    "The 'prompts_called' must be a list of 'MCPPromptCall' with result of type 'GetPromptResult' from mcp.types"
                )

        return data



try:
    import openai  # noqa: F401
except ImportError:
    raise ModuleNotFoundError(
        "Please install OpenAI to use this feature: 'pip install openai'"
    )


try:
    from openai import OpenAI, AsyncOpenAI  # noqa: F401
except ImportError:
    OpenAI = None  # type: ignore
    AsyncOpenAI = None  # type: ignore


if OpenAI or AsyncOpenAI:
    from deepeval.openai.patch import patch_openai_classes

    patch_openai_classes()



from typing import List, Optional, Dict, Union
from tqdm import tqdm

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import (
    DeepEvalBaseBenchmark,
    DeepEvalBaseBenchmarkResult,
)
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.drop.task import DROPTask
from deepeval.benchmarks.drop.template import DROPTemplate
from deepeval.benchmarks.utils import should_use_batch
from deepeval.benchmarks.schema import (
    DROPDateSchema,
    DROPNumberSchema,
    DROPStringSchema,
)
from deepeval.telemetry import capture_benchmark_run

DELIMITER = ","


class DROP(DeepEvalBaseBenchmark):
    def __init__(
        self,
        tasks: List[DROPTask] = None,
        n_shots: int = 5,
        n_problems_per_task: Optional[int] = None,
        verbose_mode: bool = False,
        **kwargs,
    ):
        from deepeval.scorer import Scorer
        import pandas as pd

        assert n_shots <= 5, "DROP only supports n_shots <= 5"
        super().__init__(**kwargs)
        self.tasks: List[DROPTask] = list(DROPTask) if tasks is None else tasks
        self.n_problems_per_task: Optional[int] = n_problems_per_task

        self.scorer = Scorer()
        self.shots_dataset: List[Dict] = None
        self.n_shots: int = n_shots
        self.predictions: Optional[pd.DataFrame] = None
        self.task_scores: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None
        self.verbose_mode: bool = verbose_mode

    def evaluate(
        self,
        model: DeepEvalBaseLLM,
        *args,
        batch_size: Union[int, None] = None,
        **kwargs,
    ) -> DeepEvalBaseBenchmarkResult:
        import pandas as pd

        with capture_benchmark_run("DROP", len(self.tasks)):
            overall_correct_predictions = 0
            overall_total_predictions = 0
            predictions_row = []
            scores_row = []
            use_batch = should_use_batch(model, batch_size)

            for task in self.tasks:
                goldens = self.load_benchmark_dataset(task)
                if (
                    self.n_problems_per_task is not None
                    and self.n_problems_per_task < len(goldens)
                ):
                    goldens = goldens[: self.n_problems_per_task]
                task_correct_predictions = 0
                task_total_predictions = len(goldens)
                overall_total_predictions += len(goldens)

                # Calculate task accuracy
                if use_batch:
                    for i in tqdm(
                        range(0, len(goldens), batch_size),
                        desc=f"Batch Processing {task.value} (batch_size={batch_size})",
                    ):
                        goldens_batch = goldens[i : i + batch_size]
                        batch_predictions = self.batch_predict(
                            model, goldens_batch
                        )
                        for golden, prediction_dict in zip(
                            goldens_batch, batch_predictions
                        ):
                            prediction = prediction_dict["prediction"]
                            score = prediction_dict["score"]
                            if score:
                                task_correct_predictions += 1
                                overall_correct_predictions += 1
                            predictions_row.append(
                                (
                                    task.value,
                                    golden.input,
                                    prediction,
                                    golden.expected_output,
                                    score,
                                )
                            )
                else:
                    for idx, golden in enumerate(
                        tqdm(goldens, desc=f"Processing {task.value}")
                    ):
                        prediction, score = self.predict(model, golden).values()
                        if score:
                            task_correct_predictions += 1
                            overall_correct_predictions += 1
                        predictions_row.append(
                            (
                                task.value,
                                golden.input,
                                prediction,
                                golden.expected_output,
                                score,
                            )
                        )
                        if self.verbose_mode:
                            self.print_verbose_logs(
                                idx,
                                task.value,
                                golden.input,
                                golden.expected_output,
                                prediction,
                                score,
                            )

                task_accuracy = (
                    task_correct_predictions / task_total_predictions
                )
                print(
                    f"DROP Task Accuracy (task={task.value}): {task_accuracy}"
                )
                scores_row.append((task.value, task_accuracy))

            # Calculate overall accuracy
            overall_accuracy = (
                overall_correct_predictions / overall_total_predictions
            )
            print(f"Overall DROP Accuracy: {overall_accuracy}")

            # Create a DataFrame from task_results_data
            # Columns: 'Task', 'Input', 'Prediction', 'Score'
            self.predictions = pd.DataFrame(
                predictions_row,
                columns=[
                    "Task",
                    "Input",
                    "Prediction",
                    "Expected Output",
                    "Correct",
                ],
            )
            self.task_scores = pd.DataFrame(
                scores_row, columns=["Task", "Score"]
            )
            self.overall_score = overall_accuracy

            return DeepEvalBaseBenchmarkResult(
                overall_accuracy=overall_accuracy
            )

    def predict(self, model: DeepEvalBaseLLM, golden: Golden) -> Dict:
        # Define prompt template
        assert (
            self.shots_dataset != None
        ), "Example dataset is empty. Call load_benchmark."
        prompt: dict = DROPTemplate.generate_output(
            train_set=self.shots_dataset,
            input=golden.input,
            n_shots=self.n_shots,
        )

        # Enforced model generation
        type_info = golden.context[0]
        try:
            if type_info == "number":
                schema = DROPNumberSchema
            elif type_info == "date":
                schema = DROPDateSchema
            elif type_info == "span":
                schema = DROPStringSchema
            res: Union[DROPNumberSchema, DROPDateSchema, DROPStringSchema] = (
                model.generate(prompt=prompt, schema=schema)
            )
            prediction = str(res.answer)
        except TypeError:
            prompt += f"Output should be a {type_info}. No explanation needed."
            prediction = model.generate(prompt)

        # For native models, shouldn't happen but just in case
        if isinstance(prediction, tuple):
            prediction = prediction[0]
        prediction = str(prediction)

        # Define Metric
        expected_output = DROPTemplate.parse_str_to_list(
            golden.expected_output, DELIMITER
        )
        score = self.scorer.quasi_contains_score(expected_output, prediction)
        return {"prediction": prediction, "score": score}

    def batch_predict(
        self, model: DeepEvalBaseLLM, goldens: List[Golden]
    ) -> List[Dict]:
        # Define prompt template
        assert (
            self.shots_dataset != None
        ), "Example dataset is empty. Call load_benchmark."

        prompts = []
        schemas = []
        for golden in goldens:
            prompt: dict = DROPTemplate.generate_output(
                train_set=self.shots_dataset,
                input=golden.input,
                type=golden.context[0],
                n_shots=self.n_shots,
            )
            prompts.append(prompt)
            output_type = golden.context[0]
            if output_type == "number":
                schema = DROPNumberSchema
            elif output_type == "date":
                schema = DROPDateSchema
            elif output_type == "span":
                schema = DROPStringSchema
            schemas.append(schema)

        try:
            responses: List[
                Union[DROPNumberSchema, DROPDateSchema, DROPStringSchema]
            ] = model.batch_generate(prompts=prompts, schemas=schemas)
            predictions = [str(res.answer) for res in responses]
        except TypeError:
            prompts = [
                prompt
                + "Output should be of type {type}. No explanation needed.".format(
                    type=type
                )
                for prompt in prompts
            ]
            predictions = model.batch_generate(prompts)

        if len(predictions) is not len(goldens):
            raise ValueError(
                "Custom `batch_generate` method did not return the same number of generations as the number of prompts."
            )

        res = []
        for i in range(len(predictions)):
            prediction = predictions[i]
            golden = goldens[i]
            # Define Metric
            score = self.scorer.quasi_exact_match_score(
                golden.expected_output, prediction
            )
            res.append({"prediction": prediction, "score": score})

        return res

    def load_benchmark_dataset(self, task: DROPTask) -> List[Golden]:
        from datasets import load_dataset

        # cache dataset
        if self.dataset:
            dataset = self.dataset
        else:
            dataset = load_dataset("ucinlp/drop")
            self.dataset = dataset

        # construct example dataset
        if not self.shots_dataset:
            train_set = dataset["train"]
            shots_set = []
            categories_seen = set()
            for data in train_set:
                category = data["section_id"]
                if category not in categories_seen:
                    categories_seen.add(category)
                    shots_set.append(data)
            self.shots_dataset = shots_set

        val_set = dataset["validation"].filter(
            lambda data: data["section_id"] == task.value
        )

        # construct test set
        goldens: List[Golden] = []
        for data in val_set:
            input = DROPTemplate.format_question(data, include_answer=False)
            output = DROPTemplate.parse_list_to_str(
                data["answers_spans"]["spans"], DELIMITER
            )
            output_type = data["answers_spans"]["types"][0]
            golden = Golden(
                input=input, expected_output=output, context=[output_type]
            )
            goldens.append(golden)

        return goldens

    def print_verbose_logs(
        self,
        idx: int,
        task_value: str,
        input: str,
        expected_output: str,
        prediction: str,
        score: int,
    ) -> str:
        steps = [
            f"Input:\n{input}",
            f"Score: {score}\nPrediction: {prediction}\nAccepted Expected Output: {expected_output}",
        ]
        verbose_logs = ""
        for i in range(len(steps) - 1):
            verbose_logs += steps[i]

            # don't add new line for penultimate step
            if i < len(steps) - 2:
                verbose_logs += " \n \n"

        if self.verbose_mode:
            print("*" * 50)
            print(f"Problem {idx + 1} (Task = {task_value})")
            print("*" * 50)
            print("")
            print(verbose_logs + f"\n \n{steps[-1]}")
            print("")
            print("=" * 70)

        return verbose_logs



"""
Applies CLI driven updates to the live Settings and optionally persists them to a
dotenv file. Also syncs os.environ, handles unsets, and warns on unknown fields.
Primary entrypoint: update_settings_and_persist.
"""

import logging
import os

from difflib import get_close_matches
from pathlib import Path
from typing import Any, Dict, Iterable, Mapping, Optional, Tuple, Union
from enum import Enum

from pydantic import SecretStr
from deepeval.config.settings import get_settings, _SAVE_RE
from deepeval.cli.dotenv_handler import DotenvHandler
from deepeval.config.utils import bool_to_env_str

logger = logging.getLogger(__name__)
StrOrEnum = Union[str, Enum]


def _env_key(k: StrOrEnum) -> str:
    return k.value if isinstance(k, Enum) else str(k)


def _normalize_for_env(val: Any) -> Optional[str]:
    """Convert typed value to string for dotenv + os.environ; None -> unset."""
    if val is None:
        return None
    if isinstance(val, SecretStr):
        return val.get_secret_value()
    if isinstance(val, bool):
        return bool_to_env_str(val)
    return str(val)


def _resolve_save_path(save_opt: Optional[str]) -> Tuple[bool, Optional[Path]]:
    """
    Returns (ok, path).
      - ok=False -> invalid save option format
      - ok=True, path=None -> no persistence requested
      - ok=True, path=Path -> persist to that file
    """
    raw = (
        save_opt if save_opt is not None else os.getenv("DEEPEVAL_DEFAULT_SAVE")
    )
    if not raw:
        return True, None
    m = _SAVE_RE.match(raw.strip())
    if not m:
        return False, None
    path = m.group("path") or ".env.local"
    path = Path(os.path.expanduser(os.path.expandvars(path)))
    return True, path


def update_settings_and_persist(
    updates: Mapping[StrOrEnum, Any],
    *,
    save: Optional[str] = None,
    unset: Iterable[StrOrEnum] = (),
    persist_dotenv: bool = True,
) -> Tuple[bool, Optional[Path]]:
    """
    Write and update:
      - validate + assign into live Settings()
      - update os.environ
      - persist to dotenv, if `save` or DEEPEVAL_DEFAULT_SAVE provided
      - unset keys where value is None or explicitly in `unset`
    Returns (handled, path_to_dotenv_if_any).
    """
    settings = get_settings()

    # validate + assign into settings.
    # validation is handled in Settings as long as validate_assignment=True
    typed: Dict[str, Any] = {}
    for key, value in updates.items():
        k = _env_key(key)
        if k not in type(settings).model_fields:
            suggestion = get_close_matches(
                k, type(settings).model_fields.keys(), n=1
            )
            if suggestion:
                logger.warning(
                    "Unknown settings field '%s'; did you mean '%s'? Ignoring.",
                    k,
                    suggestion[0],
                    stacklevel=2,
                )
            else:
                logger.warning(
                    "Unknown settings field '%s'; ignoring.", k, stacklevel=2
                )
            continue

        setattr(settings, k, value)
        # coercion is handled in Settings
        typed[k] = getattr(settings, k)

    # build env maps
    to_write: Dict[str, str] = {}
    to_unset: set[str] = set(_env_key(k) for k in unset)

    for k, v in typed.items():
        env_val = _normalize_for_env(v)
        if env_val is None:
            to_unset.add(k)
        else:
            to_write[k] = env_val

    # update process env so that it is effective immediately
    for k, v in to_write.items():
        os.environ[k] = v
    for k in to_unset:
        os.environ.pop(k, None)

    if not persist_dotenv:
        return True, None

    # persist to dotenv if save is ok
    ok, path = _resolve_save_path(save)
    if not ok:
        return False, None  # unsupported --save
    if path:
        h = DotenvHandler(path)
        if to_write:
            h.upsert(to_write)
        if to_unset:
            h.unset(to_unset)
        return True, path
    return True, None



from typing import Any, Optional, Dict
from pydantic import BaseModel


class BaseTool(BaseModel):
    name: str
    description: Optional[str] = None


class ToolSchema(BaseTool):
    parameters: Dict[str, Any]
    is_called: Optional[bool] = False


class ToolOutput(BaseTool):
    """Output of the tool function"""

    output: Any



from typing import List
from pydantic import BaseModel


class NonAdviceVerdict(BaseModel):
    verdict: str
    reason: str


class Verdicts(BaseModel):
    verdicts: List[NonAdviceVerdict]


class Advices(BaseModel):
    advices: List[str]


class NonAdviceScoreReason(BaseModel):
    reason: str



from typing import Dict, Optional, List
from openai import OpenAI, AsyncOpenAI
from deepeval.models import DeepEvalBaseEmbeddingModel
from deepeval.models.retry_policy import (
    create_retry_decorator,
    sdk_retries_for,
)
from deepeval.constants import ProviderSlug as PS


retry_openai = create_retry_decorator(PS.OPENAI)

valid_openai_embedding_models = [
    "text-embedding-3-small",
    "text-embedding-3-large",
    "text-embedding-ada-002",
]
default_openai_embedding_model = "text-embedding-3-small"


class OpenAIEmbeddingModel(DeepEvalBaseEmbeddingModel):
    def __init__(
        self,
        model: Optional[str] = None,
        _openai_api_key: Optional[str] = None,
        **kwargs,
    ):
        model_name = model if model else default_openai_embedding_model
        if model_name not in valid_openai_embedding_models:
            raise ValueError(
                f"Invalid model. Available OpenAI Embedding models: {', '.join(valid_openai_embedding_models)}"
            )
        self._openai_api_key = _openai_api_key
        self.model_name = model_name
        self.kwargs = kwargs

    @retry_openai
    def embed_text(self, text: str) -> List[float]:
        client = self.load_model(async_mode=False)
        response = client.embeddings.create(
            input=text,
            model=self.model_name,
        )
        return response.data[0].embedding

    @retry_openai
    def embed_texts(self, texts: List[str]) -> List[List[float]]:
        client = self.load_model(async_mode=False)
        response = client.embeddings.create(
            input=texts,
            model=self.model_name,
        )
        return [item.embedding for item in response.data]

    @retry_openai
    async def a_embed_text(self, text: str) -> List[float]:
        client = self.load_model(async_mode=True)
        response = await client.embeddings.create(
            input=text,
            model=self.model_name,
        )
        return response.data[0].embedding

    @retry_openai
    async def a_embed_texts(self, texts: List[str]) -> List[List[float]]:
        client = self.load_model(async_mode=True)
        response = await client.embeddings.create(
            input=texts,
            model=self.model_name,
        )
        return [item.embedding for item in response.data]

    ###############################################
    # Model
    ###############################################

    def get_model_name(self):
        return self.model_name

    def load_model(self, async_mode: bool = False):
        if not async_mode:
            return self._build_client(OpenAI)
        return self._build_client(AsyncOpenAI)

    def _client_kwargs(self) -> Dict:
        """
        If Tenacity is managing retries, force OpenAI SDK retries off to avoid double retries.
        If the user opts into SDK retries for 'openai' via DEEPEVAL_SDK_RETRY_PROVIDERS,
        leave their retry settings as is.
        """
        kwargs = dict(self.kwargs or {})
        if not sdk_retries_for(PS.OPENAI):
            kwargs["max_retries"] = 0
        return kwargs

    def _build_client(self, cls):
        kw = dict(
            api_key=self._openai_api_key,
            **self._client_kwargs(),
        )
        try:
            return cls(**kw)
        except TypeError as e:
            # older OpenAI SDKs may not accept max_retries, in that case remove and retry once
            if "max_retries" in str(e):
                kw.pop("max_retries", None)
                return cls(**kw)
            raise



from typing import Optional, Set, Any, Dict, List, Union
import threading
import asyncio
import queue
import atexit
from time import perf_counter
from enum import Enum
from pydantic import Field
from rich.console import Console

from deepeval.confident.api import Api, HttpMethods, Endpoints, is_confident
from deepeval.constants import (
    CONFIDENT_METRIC_LOGGING_FLUSH,
    CONFIDENT_METRIC_LOGGING_VERBOSE,
)
from deepeval.metrics.base_metric import BaseConversationalMetric, BaseMetric
from deepeval.test_case.conversational_test_case import ConversationalTestCase
from deepeval.test_case.llm_test_case import LLMTestCase
from deepeval.test_case.api import create_api_test_case
from deepeval.test_run.api import LLMApiTestCase, ConversationalApiTestCase
from deepeval.tracing.api import MetricData
from deepeval.config.settings import get_settings


class MetricWorkerStatus(Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    WARNING = "warning"


class ApiMetricData(MetricData):
    llm_test_case: Optional[LLMApiTestCase] = Field(None, alias="llmTestCase")
    conversational_test_case: Optional[ConversationalApiTestCase] = Field(
        None, alias="conversationalTestCase"
    )


class MetricDataManager:
    """Manager for posting metric data asynchronously in background thread."""

    def __init__(self):
        settings = get_settings()
        # Initialize queue and worker thread for metric posting
        self._metric_queue = queue.Queue()
        self._worker_thread = None
        self._min_interval = 0.2  # Minimum time between API calls (seconds)
        self._last_post_time = 0
        self._in_flight_tasks: Set[asyncio.Task[Any]] = set()
        self._flush_enabled = bool(settings.CONFIDENT_METRIC_LOGGING_FLUSH)
        self._daemon = not self._flush_enabled
        self._thread_lock = threading.Lock()
        self.metric_logging_enabled = bool(
            settings.CONFIDENT_METRIC_LOGGING_ENABLED
        )

        # Register an exit handler to warn about unprocessed metrics
        atexit.register(self._warn_on_exit)

    def post_metric_if_enabled(
        self,
        metric: Union[BaseMetric, BaseConversationalMetric],
        test_case: Optional[Union[LLMTestCase, ConversationalTestCase]] = None,
    ):
        """Post metric data asynchronously in a background thread."""
        if not self.metric_logging_enabled or not is_confident():
            return

        from deepeval.evaluate.utils import create_metric_data

        metric_data = create_metric_data(metric)
        api_metric_data = ApiMetricData(
            **metric_data.model_dump(by_alias=True, exclude_none=True)
        )

        if isinstance(test_case, LLMTestCase):
            api_metric_data.llm_test_case = create_api_test_case(test_case)
        elif isinstance(test_case, ConversationalTestCase):
            api_metric_data.conversational_test_case = create_api_test_case(
                test_case
            )

        self._ensure_worker_thread_running()
        self._metric_queue.put(api_metric_data)

    def _warn_on_exit(self):
        """Warn if there are unprocessed metrics on exit."""
        queue_size = self._metric_queue.qsize()
        in_flight = len(self._in_flight_tasks)
        remaining_tasks = queue_size + in_flight

        if not self._flush_enabled and remaining_tasks > 0:
            self._print_metric_data_status(
                metric_worker_status=MetricWorkerStatus.WARNING,
                message=f"Exiting with {queue_size + in_flight} abandoned metric(s).",
                description=f"Set {CONFIDENT_METRIC_LOGGING_FLUSH}=1 as an environment variable to flush remaining metrics to Confident AI.",
            )

    def _ensure_worker_thread_running(self):
        """Ensure the background worker thread is running."""
        with self._thread_lock:
            if (
                self._worker_thread is None
                or not self._worker_thread.is_alive()
            ):
                self._worker_thread = threading.Thread(
                    target=self._process_metric_queue,
                    daemon=self._daemon,
                )
                self._worker_thread.start()

    def _print_metric_data_status(
        self,
        metric_worker_status: MetricWorkerStatus,
        message: str,
        description: Optional[str] = None,
    ):
        """Print metric data worker status messages."""
        if getattr(get_settings(), CONFIDENT_METRIC_LOGGING_VERBOSE, False):
            console = Console()
            message_prefix = "[dim][Confident AI Metric Data Log][/dim]"
            if metric_worker_status == MetricWorkerStatus.SUCCESS:
                message = f"[green]{message}[/green]"
            elif metric_worker_status == MetricWorkerStatus.FAILURE:
                message = f"[red]{message}[/red]"
            elif metric_worker_status == MetricWorkerStatus.WARNING:
                message = f"[yellow]{message}[/yellow]"

            if bool(CONFIDENT_METRIC_LOGGING_VERBOSE):
                if description:
                    message += f": {description}"

                console.print(
                    message_prefix,
                    message,
                    f"\nTo disable dev logging, set {CONFIDENT_METRIC_LOGGING_VERBOSE}=0 as an environment variable.",
                )

    def _process_metric_queue(self):
        """Worker thread function that processes the metric queue."""
        import threading

        main_thr = threading.main_thread()

        # Create a new event loop
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # Buffer for payloads that need to be sent after main exits
        remaining_metric_request_bodies: List[Dict[str, Any]] = []

        async def _a_send_metric(metric_data: ApiMetricData):
            nonlocal remaining_metric_request_bodies
            try:
                # Build API object & payload
                try:
                    body = metric_data.model_dump(
                        by_alias=True,
                        exclude_none=True,
                    )
                except AttributeError:
                    # Pydantic version below 2.0
                    body = metric_data.dict(by_alias=True, exclude_none=True)

                # If the main thread is still alive, send now
                if main_thr.is_alive():
                    api = Api()
                    _, _ = await api.a_send_request(
                        method=HttpMethods.POST,
                        endpoint=Endpoints.METRIC_DATA_ENDPOINT,
                        body=body,
                    )
                    queue_size = self._metric_queue.qsize()
                    in_flight = len(self._in_flight_tasks)
                    status = f"({queue_size} metric{'s' if queue_size!=1 else ''} remaining in queue, {in_flight} in flight)"
                    self._print_metric_data_status(
                        metric_worker_status=MetricWorkerStatus.SUCCESS,
                        message=f"Successfully posted metric data {status}",
                    )
                elif self._flush_enabled:
                    # Main thread gone → to be flushed
                    remaining_metric_request_bodies.append(body)

            except Exception as e:
                queue_size = self._metric_queue.qsize()
                in_flight = len(self._in_flight_tasks)
                status = f"({queue_size} metric{'s' if queue_size!=1 else ''} remaining in queue, {in_flight} in flight)"
                self._print_metric_data_status(
                    metric_worker_status=MetricWorkerStatus.FAILURE,
                    message=f"Error posting metric data {status}",
                    description=str(e),
                )
            finally:
                task = asyncio.current_task()
                if task:
                    self._in_flight_tasks.discard(task)

        async def async_worker():
            # Continue while user code is running or work remains
            while (
                main_thr.is_alive()
                or not self._metric_queue.empty()
                or self._in_flight_tasks
            ):
                try:
                    metric_data = self._metric_queue.get(
                        block=True, timeout=1.0
                    )

                    # Rate-limit
                    now = perf_counter()
                    elapsed = now - self._last_post_time
                    if elapsed < self._min_interval:
                        await asyncio.sleep(self._min_interval - elapsed)
                    self._last_post_time = perf_counter()

                    # Schedule async send
                    task = asyncio.create_task(_a_send_metric(metric_data))
                    self._in_flight_tasks.add(task)
                    self._metric_queue.task_done()

                except queue.Empty:
                    await asyncio.sleep(0.1)
                    continue
                except Exception as e:
                    self._print_metric_data_status(
                        message="Error in metric worker",
                        metric_worker_status=MetricWorkerStatus.FAILURE,
                        description=str(e),
                    )
                    await asyncio.sleep(1.0)

        try:
            loop.run_until_complete(async_worker())
        finally:
            # Drain any pending tasks
            pending = asyncio.all_tasks(loop=loop)
            if pending:
                loop.run_until_complete(
                    asyncio.gather(*pending, return_exceptions=True)
                )
            self._flush_metrics(remaining_metric_request_bodies)
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    def _flush_metrics(
        self, remaining_metric_request_bodies: List[Dict[str, Any]]
    ):
        """Flush remaining metrics synchronously."""
        if not remaining_metric_request_bodies:
            return

        self._print_metric_data_status(
            MetricWorkerStatus.WARNING,
            message=f"Flushing {len(remaining_metric_request_bodies)} remaining metric(s)",
        )

        for body in remaining_metric_request_bodies:
            try:
                api = Api()
                _, link = api.send_request(
                    method=HttpMethods.POST,
                    endpoint=Endpoints.METRIC_DATA_ENDPOINT,
                    body=body,
                )
                qs = self._metric_queue.qsize()
                self._print_metric_data_status(
                    metric_worker_status=MetricWorkerStatus.SUCCESS,
                    message=f"Successfully posted metric data ({qs} metrics remaining in queue, 1 in flight)",
                    description=link,
                )
            except Exception as e:
                qs = self._metric_queue.qsize()
                self._print_metric_data_status(
                    metric_worker_status=MetricWorkerStatus.FAILURE,
                    message="Error flushing remaining metric(s)",
                    description=str(e),
                )


# Global metric manager instance
metric_data_manager = MetricDataManager()



from deepeval.test_case.llm_test_case import LLMTestCase, ToolCall
from deepeval.tracing.types import BaseSpan
from typing import Any

try:
    from llama_index.core.agent.workflow.workflow_events import (
        AgentOutput,
        AgentWorkflowStartEvent,
    )

    llama_index_agent_installed = True
except:
    llama_index_agent_installed = False


def is_llama_index_agent_installed():
    if not llama_index_agent_installed:
        raise ImportError(
            "llama-index is neccesary for this functionality. Please install it with `pip install llama-index` or with package manager of choice."
        )


def parse_id(id_: str) -> tuple[str, str]:
    """
    Parse the id_ into a tuple of class name and method name, ignoring any suffix after '-'.
    Returns empty strings as defaults if parsing fails.
    """
    try:
        # Ignore everything after the first '-'
        main_part = id_.split("-", 1)[0]
        # Split by '.' to get class and method
        parts = main_part.rsplit(".", 1)
        if len(parts) == 2:
            class_name, method_name = parts
        else:
            # If no '.' found, treat the whole string as class_name
            class_name, method_name = main_part, ""
        return class_name, method_name
    except:
        # Return empty strings if any parsing fails
        return "", ""


def prepare_input_llm_test_case_params(
    class_name: str, method_name: str, span: BaseSpan, args: dict
):

    # condition for parent agent span
    if method_name == "run":
        start_event = args.get("start_event")

        is_llama_index_agent_installed()
        if isinstance(start_event, AgentWorkflowStartEvent):
            input = ""
            for key, value in start_event.items():
                input += f"{key}: {value}\n"

            span.llm_test_case = LLMTestCase(
                input=input,
                actual_output="",
            )


def prepare_output_llm_test_case_params(
    class_name: str, method_name: str, result: Any, span: BaseSpan
):

    if class_name == "Workflow" and method_name == "run":

        is_llama_index_agent_installed()
        if isinstance(result, AgentOutput):
            span.llm_test_case.actual_output = result.response.content

            tool_calls = []
            for tool_call in result.tool_calls:
                tool_calls.append(
                    ToolCall(
                        name=tool_call.tool_name,
                        input_parameters=tool_call.tool_kwargs,
                    )
                )

            span.llm_test_case.tools_called = tool_calls



class LogiQATemplate:

    n_shot_examples = [
        "Input\nWrite a multi-choice question for the following article:\nArticle: David knows Mr. Zhang's friend Jack, and Jack knows David's friend Ms. Lin. Everyone of them who knows Jack has a master's degree, and everyone of them who knows Ms. Lin is from Shanghai.\nQuestion: \nWho is from Shanghai and has a master's degree?\nOptions:\nA David\nB Jack\nC Mr Zhang\nD Ms. Lin\nAnswer:\nA\n",
        "Input\nWrite a multi-choice question for the following article:\nArticle: Jimmy asked Hank to go to the mall the next day. Hank said, If it doesn't rain tomorrow, I'll go climbing. The next day, there was a drizzle. Jimmy thought that Hank would not go climbing, so he went to pick up Henry to the mall. Nevertheless, Hank went climbing the mountain. When the two met again, Jimmy blamed Hank for not keeping his word.\nQuestion: \nWhich of the following comments is appropriate?\nOptions:\nA This argument between Jimmy and Hank is meaningless\nB Jimmy's reasoning is illogical\nC Two people have different understandings of a drizzle\nD Hank broke his promise and caused the debate\nAnswer:\nB\n",
        "Input\nWrite a multi-choice question for the following article:\nArticle: Only if the government reinforce basic education can we improve our nation's education to a new stage. In order to stand out among other nations, we need to have a strong educational enterprise.\nQuestion: \nWhich can be inferred from the statement above?\nOptions:\nA The whole society should be focused on education\nB In order to stand out among nations, we should reinforce basic education\nC In order to improve our education to a new stage, it is necessary to increase the salary of college teachers\nD In order to reinforce basic education, all primary school teachers must have a bachelor degree or above.\nAnswer:\nB\n",
        "Input\nWrite a multi-choice question for the following article:\nArticle: Last night, Mark either went to play in the gym or visited his teacher Tony. If Mark drove last night, he didn't go to play in the gym. Mark would go visit his teacher Tony only if he and his teacher had an appointment. In fact, Mark had no appointment with his teacher Tony in advance.\nQuestion: \nWhich is true based on the above statement?\nOptions:\nA Mark went to the gym with his teacher Tony last night\nB Mark visited his teacher Tony last night\nC Mark didn't drive last night\nD Mark didn't go to the gym last night.\nAnswer:\nC\n",
        "Input\nWrite a multi-choice question for the following article:\nArticle: The coach of a national football team found that the best cooperative arrangement of the players U, V, W, X, Y, and Z during the training are: (1) V and X cannot be on the field at the same time, and neither can be off the field the same time. (2) V is not on the field only if U is not on the field. (3) If W is on the field, then X is on the field. (4) If Ｙ and Ｚ are on the field, then W must be on the field. This arrangement can yield the best performance.\nQuestion: \nIf U and Z are both on the field, for best performance, which of the following arrangement is appropriate?\nOptions:\nA X is on the eld and Y is not on the field\nB V is on the eld and Y is not on the field\nC V and W are both on the field\nD V and Y are not on the field\nAnswer:\nB\n",
    ]

    @staticmethod
    def generate_output(input: str, n_shots: int):
        prompt = ""
        for i in range(n_shots):
            prompt += LogiQATemplate.n_shot_examples[i] + "\n"
        prompt += input
        return prompt

    @staticmethod
    def format_question(data: dict):
        label_map = {0: "A", 1: "B", 2: "C", 3: "D"}
        article = data["text"]
        question = data["question"]
        options_old = data["options"]
        options = ""
        for j, option in enumerate(options_old):
            options += label_map[j] + " " + option + "\n"
        return (
            "Write a multi-choice question for the following article:\nArticle: "
            + article
            + "\nQuestion: "
            + question
            + "\nOptions: "
            + options
            + "Answer: "
        )

    @staticmethod
    def format_output(data: dict):
        label_map = {0: "A", 1: "B", 2: "C", 3: "D"}
        answer = data["answer"]
        return label_map[answer]



import functools
from deepeval.metrics import BaseMetric
from deepeval.tracing.context import current_span_context
from typing import List, Optional, Callable
from langchain_core.tools import tool as original_tool, BaseTool


def tool(
    *args,
    metrics: Optional[List[BaseMetric]] = None,
    metric_collection: Optional[str] = None,
    **kwargs
):
    """
    Patched version of langchain_core.tools.tool that prints inputs and outputs
    """

    # original_tool returns a decorator function, so we need to return a decorator
    def decorator(func: Callable) -> BaseTool:
        func = _patch_tool_decorator(func, metrics, metric_collection)
        tool_instance = original_tool(*args, **kwargs)(func)
        return tool_instance

    return decorator


def _patch_tool_decorator(
    func: Callable,
    metrics: Optional[List[BaseMetric]] = None,
    metric_collection: Optional[str] = None,
):
    original_func = func

    @functools.wraps(original_func)
    def wrapper(*args, **kwargs):
        current_span = current_span_context.get()
        current_span.metrics = metrics
        current_span.metric_collection = metric_collection
        res = original_func(*args, **kwargs)
        return res

    tool = wrapper
    return tool



from typing import Literal, Dict, Any
from .base import BaseMessage


class TextMessage(BaseMessage):
    type: Literal["text", "thinking"]
    content: str


class ToolCallMessage(BaseMessage):
    """This is a message for tool calls in response.choices[0].message.tool_calls"""

    name: str
    args: Dict[str, Any]



from .azure_embedding_model import AzureOpenAIEmbeddingModel
from .openai_embedding_model import OpenAIEmbeddingModel
from .local_embedding_model import LocalEmbeddingModel
from .ollama_embedding_model import OllamaEmbeddingModel



from __future__ import annotations

from dataclasses import dataclass
from typing import Generic, TypeVar, List

from deepeval.prompt import Prompt
from deepeval.metrics import BaseMetric
from deepeval.tracing.types import LlmSpan

try:
    from agents.agent import Agent as BaseAgent
    from deepeval.openai_agents.patch import (
        patch_default_agent_runner_get_model,
    )
except Exception as e:
    raise RuntimeError(
        "openai-agents is required for this integration. Please install it."
    ) from e

TContext = TypeVar("TContext")


@dataclass
class DeepEvalAgent(BaseAgent[TContext], Generic[TContext]):
    """
    A subclass of agents.Agent.
    """

    llm_metric_collection: str = None
    llm_metrics: List[BaseMetric] = None
    confident_prompt: Prompt = None
    agent_metrics: List[BaseMetric] = None
    agent_metric_collection: str = None

    def __post_init__(self):
        patch_default_agent_runner_get_model()



from enum import Enum
from typing import Optional

# ANSI escape codes for coloring
RESET = "\033[0m"
DIM = "\033[2m"
GREEN = "\033[32m"
RED = "\033[31m"
YELLOW = "\033[33m"


class SynthesizerStatus(Enum):
    SUCCESS = "success"
    FAILURE = "failure"
    WARNING = "warning"


def print_synthesizer_status(
    trace_worker_status: SynthesizerStatus,
    message: str,
    description: Optional[str] = None,
):
    prefix = f"{DIM}[Confident AI Synthesizer Log]{RESET}"
    if trace_worker_status == SynthesizerStatus.SUCCESS:
        colored_msg = f"{GREEN}{message}{RESET}"
        label = "SUCCESS"
    elif trace_worker_status == SynthesizerStatus.FAILURE:
        colored_msg = f"{RED}{message}{RESET}"
        label = "FAILURE"
    elif trace_worker_status == SynthesizerStatus.WARNING:
        colored_msg = f"{YELLOW}{message}{RESET}"
        label = "WARNING"
    if description:
        print(f"{prefix} {label}: {colored_msg}: {description}")
    else:
        print(f"{prefix} {label}: {colored_msg}")



import asyncio
import inspect
import json
import re

from typing import List, Optional, Any
from opentelemetry.trace import Tracer

from deepeval.dataset.api import Golden
from deepeval.dataset.golden import ConversationalGolden
from deepeval.test_case import LLMTestCase, ConversationalTestCase, Turn


def convert_test_cases_to_goldens(
    test_cases: List[LLMTestCase],
) -> List[Golden]:
    goldens = []
    for test_case in test_cases:
        golden = {
            "input": test_case.input,
            "actual_output": test_case.actual_output,
            "expected_output": test_case.expected_output,
            "context": test_case.context,
            "retrieval_context": test_case.retrieval_context,
            "tools_called": test_case.tools_called,
            "expected_tools": test_case.expected_tools,
        }
        goldens.append(Golden(**golden))
    return goldens


def convert_goldens_to_test_cases(
    goldens: List[Golden],
    _alias: Optional[str] = None,
    _id: Optional[str] = None,
) -> List[LLMTestCase]:
    test_cases = []
    for index, golden in enumerate(goldens):
        test_case = LLMTestCase(
            input=golden.input,
            actual_output=golden.actual_output,
            expected_output=golden.expected_output,
            context=golden.context,
            retrieval_context=golden.retrieval_context,
            tools_called=golden.tools_called,
            expected_tools=golden.expected_tools,
            name=golden.name,
            comments=golden.comments,
            additional_metadata=golden.additional_metadata,
            _dataset_alias=_alias,
            _dataset_id=_id,
            _dataset_rank=index,
        )
        test_cases.append(test_case)
    return test_cases


def convert_convo_test_cases_to_convo_goldens(
    test_cases: List[ConversationalTestCase],
) -> List[ConversationalGolden]:
    goldens = []
    for test_case in test_cases:
        if not test_case.scenario:
            raise ValueError(
                "Please provide a scenario in your 'ConversationalTestCase' to convert it to a 'ConversationalGolden'."
            )
        golden = {
            "scenario": test_case.scenario,
            "turns": test_case.turns,
            "expected_outcome": test_case.expected_outcome,
            "user_description": test_case.user_description,
            "context": test_case.context,
        }
        goldens.append(ConversationalGolden(**golden))
    return goldens


def convert_convo_goldens_to_convo_test_cases(
    goldens: List[ConversationalGolden],
    _alias: Optional[str] = None,
    _id: Optional[str] = None,
) -> List[ConversationalTestCase]:
    test_cases = []
    for index, golden in enumerate(goldens):
        test_case = ConversationalTestCase(
            turns=golden.turns or [],
            scenario=golden.scenario,
            user_description=golden.user_description,
            context=golden.context,
            name=golden.name,
            additional_metadata=golden.additional_metadata,
            comments=golden.comments,
            _dataset_alias=_alias,
            _dataset_id=_id,
            _dataset_rank=index,
        )
        test_cases.append(test_case)
    return test_cases


def trimAndLoadJson(input_string: str) -> Any:
    try:
        cleaned_string = re.sub(r",\s*([\]}])", r"\1", input_string.strip())
        return json.loads(cleaned_string)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON: {input_string}. Error: {str(e)}")
    except Exception as e:
        raise Exception(f"An unexpected error occurred: {str(e)}")


def format_turns(turns: List[Turn]) -> str:
    res = []
    for turn in turns:
        cur_turn = {
            "role": turn.role,
            "content": turn.content,
            "retrieval_context": (
                turn.retrieval_context if turn.retrieval_context else None
            ),
        }
        res.append(cur_turn)
    try:
        return json.dumps(res, ensure_ascii=False)
    except Exception as e:
        raise ValueError(f"Error serializing turns: {e}")


def parse_turns(turns_str: str) -> List[Turn]:
    try:
        parsed = json.loads(turns_str)
    except json.JSONDecodeError as e:
        raise ValueError(f"Invalid JSON: {e}")

    if not isinstance(parsed, list):
        raise TypeError("Expected a list of turns.")

    res = []
    for i, turn in enumerate(parsed):
        if not isinstance(turn, dict):
            raise TypeError(f"Turn at index {i} is not a dictionary.")

        # Ensuring 'role' and 'content' are strings
        if "role" not in turn or not isinstance(turn["role"], str):
            raise ValueError(f"Turn at index {i} is missing a valid 'role'.")
        if "content" not in turn or not isinstance(turn["content"], str):
            raise ValueError(f"Turn at index {i} is missing a valid 'content'.")

        retrieval_context = turn.get("retrieval_context")

        res.append(
            Turn(
                role=turn["role"],
                content=turn["content"],
                retrieval_context=retrieval_context,
            )
        )
    return res


def check_tracer(tracer: Optional[Tracer] = None) -> Tracer:
    if tracer:
        return tracer
    # Prefer module-level test-run tracer if available
    try:
        from deepeval.dataset.test_run_tracer import (
            GLOBAL_TEST_RUN_TRACER,
        )

        if GLOBAL_TEST_RUN_TRACER is not None:
            return GLOBAL_TEST_RUN_TRACER
    except Exception:
        raise RuntimeError(
            "No global OpenTelemetry tracer provider is configured."  # TODO: link to docs
        )

    return GLOBAL_TEST_RUN_TRACER


def coerce_to_task(obj: Any) -> asyncio.Future[Any]:
    # already a Task so just return it
    if isinstance(obj, asyncio.Task):
        return obj

    # If it is a future, it is already scheduled, so just return it
    if asyncio.isfuture(obj):
        # type: ignore[return-value]  # it is an awaitable, gather accepts it
        return obj

    # bare coroutine must be explicitly scheduled using create_task to bind to loop & track
    if asyncio.iscoroutine(obj):
        return asyncio.create_task(obj)

    # generic awaitable (any object with __await__) will need to be wrapped so create_task accepts it
    if inspect.isawaitable(obj):

        async def _wrap(awaitable):
            return await awaitable

        return asyncio.create_task(_wrap(obj))

    # not awaitable, so time to sound the alarm!
    raise TypeError(
        f"Expected Task/Future/coroutine/awaitable, got {type(obj).__name__}"
    )



class BoolQTemplate:

    n_shot_examples = [
        "Q: do iran and afghanistan speak the same language?\nP: Persian (/ˈpɜːrʒən, -ʃən/), also known by its endonym Farsi (فارسی fārsi (fɒːɾˈsiː) ( listen)), is one of the Western Iranian languages within the Indo-Iranian branch of the Indo-European language family. It is primarily spoken in Iran, Afghanistan (officially known as Dari since 1958), and Tajikistan (officially known as Tajiki since the Soviet era), and some other regions which historically were Persianate societies and considered part of Greater Iran. It is written in the Persian alphabet, a modified variant of the Arabic script, which itself evolved from the Aramaic alphabet.\nA: Yes\n",
        "Q: is elder scrolls online the same as skyrim?\nP: As with other games in The Elder Scrolls series, the game is set on the continent of Tamriel. The events of the game occur a millennium before those of The Elder Scrolls V: Skyrim and around 800 years before The Elder Scrolls III: Morrowind and The Elder Scrolls IV: Oblivion. It has a broadly similar structure to Skyrim, with two separate conflicts progressing at the same time, one with the fate of the world in the balance, and one where the prize is supreme power on Tamriel. In The Elder Scrolls Online, the first struggle is against the Daedric Prince Molag Bal, who is attempting to meld the plane of Mundus with his realm of Coldharbour, and the second is to capture the vacant imperial throne, contested by three alliances of the mortal races. The player character has been sacrificed to Molag Bal, and Molag Bal has stolen their soul, the recovery of which is the primary game objective.\nA: No\n",
        "Q: do good samaritan laws protect those who help at an accident?\nP: Good Samaritan laws offer legal protection to people who give reasonable assistance to those who are, or who they believe to be, injured, ill, in peril, or otherwise incapacitated. The protection is intended to reduce bystanders' hesitation to assist, for fear of being sued or prosecuted for unintentional injury or wrongful death. An example of such a law in common-law areas of Canada: a good Samaritan doctrine is a legal principle that prevents a rescuer who has voluntarily helped a victim in distress from being successfully sued for wrongdoing. Its purpose is to keep people from being reluctant to help a stranger in need for fear of legal repercussions should they make some mistake in treatment. By contrast, a duty to rescue law requires people to offer assistance and holds those who fail to do so liable.\nA: Yes\n",
        "Q: is windows movie maker part of windows essentials?\nP: Windows Movie Maker (formerly known as Windows Live Movie Maker in Windows 7) is a discontinued video editing software by Microsoft. It is a part of Windows Essentials software suite and offers the ability to create and edit videos as well as to publish them on OneDrive, Facebook, Vimeo, YouTube, and Flickr.\nA: Yes\n",
        "Q: can you use oyster card at epsom station?\nP: Epsom railway station serves the town of Epsom in Surrey. It is located off Waterloo Road and is less than two minutes' walk from the High Street. It is not in the London Oyster card zone unlike Epsom Downs or Tattenham Corner stations. The station building was replaced in 2012/2013 with a new building with apartments above the station (see end of article).\nA: No\n",
    ]

    @staticmethod
    def generate_output(input: str, n_shots: int):
        prompt = ""
        for i in range(n_shots):
            prompt += BoolQTemplate.n_shot_examples[i] + "\n"
        prompt += input
        return prompt

    @staticmethod
    def format_question(data: dict):
        question = data["question"]
        passage = data["passage"]
        prompt = ""
        prompt += f"Q: {question}?\n"
        prompt += f"P: {passage}\n"
        prompt += "A: "
        return prompt

    @staticmethod
    def format_answer(data: dict):
        answer = data["answer"]
        return "Yes" if answer else "No"
